# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Predictor class to handle detection of poses for movenet
"""

import logging
from pathlib import Path
from typing import Any, Dict, List, Tuple

import cv2 as cv
import numpy as np
import tensorflow as tf
from tensorflow.python.saved_model import tag_constants

# fmt: off
SKELETON = [
    [16, 14], [14, 12], [17, 15], [15, 13], [12, 13], [6, 12], [7, 13], [6, 7],
    [6, 8], [7, 9], [8, 10], [9, 11], [2, 3], [1, 2], [1, 3], [2, 4], [3, 5],
    [4, 6], [5, 7],
]
# fmt: on


class Predictor:
    """Predictor class to handle detection of poses for MoveNet."""

    def __init__(self, config: Dict[str, Any], model_dir: Path) -> None:
        self.logger = logging.getLogger(__name__)
        self.config = config
        self.model_dir = model_dir
        self.model_type = self.config["model_type"]
        self.movenet_model = self._create_movenet_model()

    def _create_movenet_model(self) -> tf.keras.Model:
        """
        Loads the MoveNet model

        Returns:
            model (tf.keras.Model): Initialized MoveNet Model
        """
        model_path = (
            self.model_dir / self.config["weights"]["model_file"][self.model_type]
        )
        model = tf.saved_model.load(str(model_path), tags=[tag_constants.SERVING])
        self.resolution = self.get_resolution_as_tuple(
            self.config["resolution"][self.model_type]
        )

        if "multi" in self.model_type:
            self.logger.info(
                (
                    f"MoveNet model loaded with following configs: \n\t"
                    f"Model type: {self.model_type}, \n\t"
                    f"Input resolution: {self.resolution}, \n\t"
                    f"bbox_score_threshold: {self.config['bbox_score_threshold']}, \n\t"
                    f"keypoint_score_threshold: {self.config['keypoint_score_threshold']}"
                )
            )
        else:
            # movenet singlepose do not output bbox, so bbox score
            # threshold not applicable
            self.logger.info(
                (
                    f"MoveNet model loaded with following configs: \n\t"
                    f"Model type: {self.model_type}, \n\t"
                    f"Input resolution: {self.resolution}, \n\t"
                    f"bbox_score_threshold: NA for singlepose models, \n\t"
                    f"keypoint_score_threshold: {self.config['keypoint_score_threshold']}"
                ),
            )

        return model

    @staticmethod
    def get_resolution_as_tuple(resolution: dict) -> Tuple[int, int]:
        """Convert resolution from dict to tuple format

        Args:
            resolution (dict): height and width in dict format

        Returns:
            resolution (Tuple(int)): height and width in tuple format
        """
        res1, res2 = resolution["height"], resolution["width"]

        return int(res1), int(res2)

    def predict(
        self, frame: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        # pylint: disable=too-many-locals
        """MoveNet prediction function

        Args:
            frame (np.ndarray): image in numpy array

        Returns:
            bboxes (np.ndarray): Nx4 array of bboxes, N is number of detections
            keypoints (np.ndarray): Nx17x2 array of keypoint coordinates
            keypoints_scores (np.ndarray): Nx17 array of keypoint scores
            keypoints_conns (np.ndarray): NxD'x2 keypoint connections, where
                D' is the varying pairs of valid keypoint connections per detection
        """
        image_data = cv.resize(frame, (self.resolution))
        image_data = cv.cvtColor(image_data, cv.COLOR_BGR2RGB)
        image_data = np.asarray([image_data]).astype(np.int32)
        infer = self.movenet_model.signatures["serving_default"]
        outputs = infer(tf.constant(image_data))
        predictions = outputs["output_0"]

        if "multi" in self.config["model_type"]:
            (
                bboxes,
                keypoints,
                keypoints_scores,
                keypoints_conns,
            ) = self._get_results_multi(predictions)
        else:
            (
                bboxes,
                keypoints,
                keypoints_scores,
                keypoints_conns,
            ) = self._get_results_single(predictions)

        return bboxes, keypoints, keypoints_scores, keypoints_conns

    def _get_results_single(
        self, predictions: tf.Tensor
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Returns formatted outputs for singlepose model
        Predictions come in a [1x1x17x3] tensor
        First 2 channel in the last dimension represent the yx coordinates
        3rd channel represents confidence scores for the keypoints

        Args:
            predictions (tf.Tensor): model output in a [1x1x17x3] tensor

        Returns:
            bboxes (np.ndarray): 1x4 array of bboxes
            keypoints (np.ndarray): 1x17x2 array of keypoint coordinates
            keypoints_scores (np.ndarray): 1x17 array of keypoint scores
            keypoints_conns (np.ndarray): 1xD'x2 keypoint connections, where
                D' is the varying pair of valid keypoint connections per detection
        """
        predictions = tf.squeeze(predictions, axis=0)
        predictions = tf.squeeze(predictions, axis=0).numpy()
        keypoints_x = predictions[:, 1]
        keypoints_y = predictions[:, 0]
        keypoints_scores = predictions[:, 2].reshape((1, -1))

        # to concatenate the xy coordinate together for each keypoint
        keypoints = (
            tf.concat(
                [
                    tf.expand_dims(keypoints_x, axis=1),
                    tf.expand_dims(keypoints_y, axis=1),
                ],
                1,
            )
        ).numpy()
        keypoints = keypoints.reshape((1, -1, 2))
        valid_keypoints, keypoints_masks = self._get_keypoints_coords(
            keypoints, keypoints_scores, self.config["keypoint_score_threshold"]
        )
        # if valid_keypoints is None:
        #     return (np.zeros(0), np.zeros(0), np.zeros(0), np.zeros(0))
        # if all the keypoints_score are below threshold, valid_keypoints
        # will be an array with all [-1,-1]
        if np.all(valid_keypoints == np.asarray([[-1, -1]])):
            return (np.zeros(0), np.zeros(0), np.zeros(0), np.zeros(0))

        keypoints_conns = self._get_connections_of_poses(keypoints, keypoints_masks)
        bbox = self._get_bbox_from_keypoints(valid_keypoints)

        return bbox, valid_keypoints, keypoints_scores, keypoints_conns

    def _get_results_multi(
        self, predictions: tf.Tensor
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Returns formatted outputs for multipose model
        Predictions come in a [1x6x56] tensor for up to 6 ppl detections
        First 17 x 3 = 51 elements are keypoints locations in the form of
        [y_0, x_0, s_0, y_1, x_1, s_1, …, y_16, x_16, s_16],
        where y_i, x_i, s_i are the yx-coordinates and confidence score
        Remaining 5 elements [ymin, xmin, ymax, xmax, score] represent
        bbox coordinates and confidence score

        Args:
            predictions (tf.Tensor): model output in a [1x6x56] tensor

        Returns:
            bboxes (np.ndarray): Nx4 array of bboxes, N is number of detections
            keypoints (np.ndarray): Nx17x2 array of keypoint coordinates
            keypoints_scores (np.ndarray): Nx17 array of keypoint scores
            keypoints_conns (np.ndarray): NxD'x2 keypoint connections, where
                D' is the varying pair of valid keypoint connections per detection
        """

        predictions = tf.squeeze(predictions, axis=0).numpy()
        keypoints_x = predictions[:, range(1, 51, 3)]
        keypoints_y = predictions[:, range(0, 51, 3)]
        keypoints_scores = predictions[:, range(2, 51, 3)]
        bboxes = predictions[:, 51:55]
        bbox_score = predictions[:, 55]
        keypoints = (
            tf.concat(
                [
                    tf.expand_dims(keypoints_x, axis=2),
                    tf.expand_dims(keypoints_y, axis=2),
                ],
                2,
            )
        ).numpy()
        # swap bbox coordinates from y1,x1,y2,x2 to x1,y1,x2,y2
        bboxes[:, [0, 1]] = bboxes[:, [1, 0]]
        bboxes[:, [2, 3]] = bboxes[:, [3, 2]]
        bbox_masks = self._get_masks_from_bbox_scores(
            bbox_score, self.config["bbox_score_threshold"]
        )
        if len(bbox_masks) == 0:
            return (np.zeros(0), np.zeros(0), np.zeros(0), np.zeros(0))

        bboxes = bboxes[bbox_masks, :]
        bbox_score = bbox_score[bbox_masks]
        keypoints = keypoints[bbox_masks, :, :]
        keypoints_scores = keypoints_scores[bbox_masks, :]
        valid_keypoints, keypoints_masks = self._get_keypoints_coords(
            keypoints, keypoints_scores, self.config["keypoint_score_threshold"]
        )
        keypoints_conns = self._get_connections_of_poses(keypoints, keypoints_masks)

        return bboxes, valid_keypoints, keypoints_scores, keypoints_conns

    @staticmethod
    def _get_connections_of_poses(
        keypoints: np.ndarray, masks: np.ndarray
    ) -> np.ndarray:
        """
        Get connections between adjacent keypoint pairs if both keypoints are detected
        Output will be in the shape of NxD'x2 keypoint connections, where D' is the
        varying pairs of valid keypoint connections per detection

        Args:
            keypoints (np.ndarray): Nx17x2 array of keypoint coordinates, where N is
                the number of detections
            masks (np.ndarray): Nx17 boolean for valid keypoints

        Returns:
            keypoint_conns (np.ndarray): NxD'x2 keypoint connections, where
                D' is the varying pair of valid keypoint connections per detection
        """
        all_connections = []
        for i in range(keypoints.shape[0]):
            connections = []
            for start_joint, end_joint in SKELETON:
                if masks[i, start_joint - 1] and masks[i, end_joint - 1]:
                    connections.append(
                        np.vstack(
                            (keypoints[i, start_joint - 1], keypoints[i, end_joint - 1])
                        )
                    )
            all_connections.append(connections)
        keypoint_conns = np.asarray(all_connections)

        return keypoint_conns

    @staticmethod
    def _get_keypoints_coords(
        keypoints: np.ndarray,
        keypoints_score: np.ndarray,
        keypoints_score_threshold: float,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Keep only valid keypoints' relative coordinates

        Args:
            keypoints (np.ndarray): Nx17x2 array of keypoints' relative coordinates
            keypoints_score (np.ndarray): Nx17 keypoints confidence score
                for valid keypoints coordinates
            keypoints_score_threshold (float): threshold for keypoints score

        Returns:
            valid_keypoints (np.ndarray): Nx17x2 array of keypoints where undetected
                keypoints are assigned a (-1,-1) value.
            keypoint_masks (np.ndarray) : Nx17 boolean for valid keypoints
        """
        valid_keypoints = keypoints.copy()
        valid_keypoints = np.clip(valid_keypoints, 0, 1)
        keypoint_masks = keypoints_score > keypoints_score_threshold
        valid_keypoints[~keypoint_masks] = [-1, -1]

        return valid_keypoints, keypoint_masks

    @staticmethod
    def _get_masks_from_bbox_scores(
        bbox_scores: np.ndarray, score_threshold: float
    ) -> List:
        """
        Obtain masks for bbox with confidence scores above the detection threshold
        The mask come in the format of a list with idx of approved detection
        eg [0,1,2], generally movenet will have the valid detections in the first few indexes

        Args:
            bbox_scores (np.ndarray): array of bbox confidence scores
            score_threshold (float): threshold value for bbox confidence scores

        Returns:
            masks (list): list of index of bboxes that have respective bbox confidence
            scores above the score threshold
        """
        bbox_scores_list = bbox_scores.tolist()
        masks = [
            idx for idx, score in enumerate(bbox_scores_list) if score > score_threshold
        ]
        return masks

    @staticmethod
    def _get_bbox_from_keypoints(valid_keypoints: np.ndarray) -> np.ndarray:
        """
        Obtain coordinates from the keypoints for single pose model where bounding box
        coordinates are not provided as model outputs. The bounding box coordinates
        will be [xmin,ymin,xmax,ymax], derived from the keypoints coordinates.

        Args:
            valid_keypoints (np.ndarray): Nx17x2 array of keypoints where undetected
                Where N is the number of detections

        Returns:
            bboxes (np.ndarray): Nx4 array of bbox coordinates in the form of
                [x1,y1,x2,y2] where the values are in relative coordinates
                between 0 and 1
        """
        # using absolute because it is coded that invalid keypoints are set as -1
        xmin = np.abs(valid_keypoints[:, :, 0]).min()
        ymin = np.abs(valid_keypoints[:, :, 1]).min()
        xmax = valid_keypoints[:, :, 0].max()
        ymax = valid_keypoints[:, :, 1].max()
        bboxes = np.asarray([[xmin, ymin, xmax, ymax]])

        return bboxes



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Functions for drawing bounding box related UI components."""

from typing import List, Tuple

import cv2
import numpy as np
from cv2 import FONT_HERSHEY_SIMPLEX, LINE_AA

from peekingduck.pipeline.nodes.draw.utils.constants import (
    BLACK,
    CHAMPAGNE,
    FILLED,
    NORMAL_FONTSCALE,
    POINT_RADIUS,
    PRIMARY_PALETTE,
)
from peekingduck.pipeline.nodes.draw.utils.constants import (
    PRIMARY_PALETTE_LENGTH as TOTAL_COLORS,
)
from peekingduck.pipeline.nodes.draw.utils.constants import THICK, VERY_THICK
from peekingduck.pipeline.nodes.draw.utils.general import (
    get_image_size,
    project_points_onto_original_image,
)


def draw_bboxes(
    frame: np.ndarray,
    bboxes: List[List[float]],
    bbox_labels: List[str],
    show_labels: bool,
    color_choice: Tuple[int, int, int] = None,
) -> None:
    """Draws bboxes onto an image frame.

    Args:
        frame (np.ndarray): Image of current frame.
        bboxes (List[List[float]]): Bounding box coordinates.
        color (Tuple[int, int, int]): Color used for bounding box.
        bbox_labels (List[str]): Labels of object detected.
    """
    image_size = get_image_size(frame)
    # Get unique label color indexes
    color_idx = {label: idx for idx, label in enumerate(set(bbox_labels))}

    for i, bbox in enumerate(bboxes):
        if color_choice:
            color = color_choice
        else:
            color = PRIMARY_PALETTE[color_idx[bbox_labels[i]] % TOTAL_COLORS]
        if show_labels:
            _draw_bbox(frame, bbox, image_size, color, bbox_labels[i])
        else:
            _draw_bbox(frame, bbox, image_size, color)


def _draw_bbox(
    frame: np.ndarray,
    bbox: np.ndarray,
    image_size: Tuple[int, int],
    color: Tuple[int, int, int],
    bbox_label: str = None,
) -> None:
    """Draws a single bounding box."""
    top_left, bottom_right = project_points_onto_original_image(bbox, image_size)
    cv2.rectangle(
        frame,
        (top_left[0], top_left[1]),
        (bottom_right[0], bottom_right[1]),
        color,
        VERY_THICK,
    )

    if bbox_label:
        _draw_label(frame, top_left, bbox_label, color, BLACK)


def _draw_label(
    frame: np.ndarray,
    top_left: Tuple[int, int],
    bbox_label: str,
    bg_color: Tuple[int, int, int],
    text_color: Tuple[int, int, int],
) -> None:
    """Draws bbox label at top left of bbox."""
    # get label size
    (text_width, text_height), baseline = cv2.getTextSize(
        bbox_label, FONT_HERSHEY_SIMPLEX, NORMAL_FONTSCALE, THICK
    )
    # put filled text rectangle
    cv2.rectangle(
        frame,
        (top_left[0], top_left[1]),
        (top_left[0] + text_width, top_left[1] - text_height - baseline),
        bg_color,
        FILLED,
    )

    # put text above rectangle
    bbox_label = bbox_label[:1].capitalize() + bbox_label[1:]
    cv2.putText(
        frame,
        bbox_label,
        (top_left[0], top_left[1] - 6),
        FONT_HERSHEY_SIMPLEX,
        NORMAL_FONTSCALE,
        text_color,
        THICK,
        LINE_AA,
    )


def draw_tags(
    frame: np.ndarray,
    bboxes: np.ndarray,
    tags: List[str],
    color: Tuple[int, int, int],
) -> None:
    """Draw tags above bboxes.

    Args:
        frame (np.ndarray): Image of current frame.
        bboxes (np.ndarray): Bounding box coordinates.
        tags (Union[List[str], List[int]]): Tag associated with bounding box.
        color (Tuple[int, int, int]): Color of text.
    """
    image_size = get_image_size(frame)
    for idx, bbox in enumerate(bboxes):
        _draw_tag(frame, bbox, tags[idx], image_size, color)


def _draw_tag(
    frame: np.ndarray,
    bbox: np.ndarray,
    tag: str,
    image_size: Tuple[int, int],
    color: Tuple[int, int, int],
) -> None:
    """Draws a tag above a single bounding box."""
    top_left, btm_right = project_points_onto_original_image(bbox, image_size)

    # Find offset to centralize text
    (text_width, _), baseline = cv2.getTextSize(
        tag, FONT_HERSHEY_SIMPLEX, NORMAL_FONTSCALE, THICK
    )
    bbox_width = btm_right[0] - top_left[0]
    offset = int((bbox_width - text_width) / 2)
    position = (top_left[0] + offset, top_left[1] - baseline)

    cv2.putText(
        frame, tag, position, FONT_HERSHEY_SIMPLEX, NORMAL_FONTSCALE, color, VERY_THICK
    )


def draw_pts(frame: np.ndarray, pts: List[Tuple[float]]) -> None:
    """Draw pts of selected object onto frame.

    Args:
        frame (np.array): Image of current frame.
        pts (List[Tuple[float]]): Bottom midpoints of bboxes.
    """
    for point in pts:
        cv2.circle(frame, point, POINT_RADIUS, CHAMPAGNE, -1)


def check_bgr_type(colors: List[int]) -> None:
    """Check the type and range of provided colors.

    Args:
        colors (List[int]): Color in BGR format.
    """
    for color in colors:
        if not isinstance(color, int):
            raise TypeError(
                f"Color values should be integers. The chosen value of: {color} is of type: "
                f"{type(color)} instead."
            )
        if color < 0 or color > 255:
            raise ValueError(
                f"Color values should lie between (and include) 0 and 255. The chosen value of: "
                f"{color} is not within this range."
            )



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Face detection class using mtcnn model to find face bboxes
"""

import logging
from pathlib import Path
from typing import Any, Dict, Tuple

import numpy as np
import tensorflow as tf

from peekingduck.pipeline.nodes.model.mtcnnv1.mtcnn_files.graph_functions import (
    load_graph,
)
from peekingduck.pipeline.utils.bbox.transforms import xyxy2xyxyn


class Detector:  # pylint: disable=too-many-instance-attributes
    """Face detection class using MTCNN model to find bboxes and landmarks"""

    def __init__(self, config: Dict[str, Any], model_dir: Path) -> None:
        self.logger = logging.getLogger(__name__)

        self.config = config
        self.model_dir = model_dir
        self.min_size = self.config["min_size"]
        self.factor = self.config["scale_factor"]
        self.thresholds = [
            float(threshold) for threshold in self.config["network_thresholds"]
        ]
        self.score = self.config["score_threshold"]
        self.mtcnn = self._create_mtcnn_model()

    def _create_mtcnn_model(self) -> tf.keras.Model:
        """
        Creates MTCNN model for face detection
        """
        model_path = self.model_dir / self.config["weights"]["model_file"]

        self.logger.info(
            "MTCNN model loaded with following configs: \n\t"
            f"Min size: {self.config['min_size']}, \n\t"
            f"Scale Factor: {self.config['scale_factor']}, \n\t"
            f"Network Thresholds: {self.config['network_thresholds']}, \n\t"
            f"Score Threshold: {self.config['score_threshold']}"
        )

        return self._load_mtcnn_graph(model_path)

    @staticmethod
    def _load_mtcnn_graph(model_path: Path) -> tf.compat.v1.GraphDef:
        if model_path.is_file():
            return load_graph(str(model_path))

        raise ValueError(
            f"Graph file does not exist. Please check that {model_path} exists"
        )

    def predict_bbox_landmarks(
        self, image: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Predicts face bboxes, scores and landmarks

        Args:
            image (np.ndarray): image in numpy array

        Returns:
            bboxes (np.ndarray): numpy array of detected bboxes
            scores (np.ndarray): numpy array of confidence scores
            landmarks (np.ndarray): numpy array of facial landmarks
            labels (np.ndarray): numpy array of class labels (i.e. face)
        """
        # 1. process inputs
        image = self.process_image(image)

        # 2. evaluate image
        bboxes, scores, landmarks = self.mtcnn(
            image, self.min_size, self.factor, self.thresholds
        )

        # 3. process outputs
        bboxes, scores, landmarks = self.process_outputs(
            image, bboxes, scores, landmarks
        )

        # 4. create bbox_labels
        classes = np.array(["face"] * len(bboxes))

        return bboxes, scores, landmarks, classes

    @staticmethod
    def process_image(image: np.ndarray) -> tf.Tensor:
        """Processes input image

        Args:
            image (np.ndarray): image in numpy array

        Returns:
            image (np.ndarray): processed numpy array of image
        """
        image = image.astype(np.float32)
        image = tf.convert_to_tensor(image)

        return image

    def process_outputs(
        self,
        image: np.ndarray,
        bboxes: tf.Tensor,
        scores: tf.Tensor,
        landmarks: tf.Tensor,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """Processes MTCNN model outputs

        Args:
            image (np.ndarray): image in numpy array
            bboxes (tf.Tensor): tensor array of detected bboxes
            scores (tf.Tensor): tensor array of confidence scores
            landmarks (tf.Tensor): tensor array of facial landmarks

        Returns:
            bboxes (np.ndarray): processed numpy array of detected bboxes
            scores (np.ndarray): processed numpy array of confidence scores
            landmarks (np.ndarray): processed numpy array of facial landmarks
        """
        bboxes, scores, landmarks = bboxes.numpy(), scores.numpy(), landmarks.numpy()

        # Filter bboxes by confidence score
        indices = np.where(scores > self.score)[0]
        bboxes = bboxes[indices]
        scores = scores[indices]
        landmarks = landmarks[indices]

        # Swap position of x, y coordinates
        bboxes[:, [0, 1]] = bboxes[:, [1, 0]]
        bboxes[:, [2, 3]] = bboxes[:, [3, 2]]

        bboxes = xyxy2xyxyn(bboxes, image.shape[0], image.shape[1])

        return bboxes, scores, landmarks



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Main engine for PeekingDuck processes.
"""

import copy
import logging
import sys
from pathlib import Path
from time import perf_counter
from typing import List

from peekingduck.declarative_loader import DeclarativeLoader, NodeList
from peekingduck.pipeline.nodes.node import AbstractNode
from peekingduck.pipeline.pipeline import Pipeline
from peekingduck.utils.requirement_checker import RequirementChecker


class Runner:
    """The runner class for creation of pipeline using declared/given nodes.

    The runner class uses the provided configurations to setup a node pipeline
    which is used to run inference.

    Args:
        pipeline_path (:obj:`pathlib.Path` | :obj:`None`): If a path to
            *pipeline_config.yml* is provided, uses
            :py:class:`DeclarativeLoader <peekingduck.declarative_loader.DeclarativeLoader>`
            to load the YAML file according to PeekingDuck's specified schema
            to obtain the declared nodes that would be sequentially initialized
            and used to create the pipeline for running inference.
        config_updates_cli (:obj:`str` | :obj:`None`): Configuration changes
            passed as part of the CLI command used to modify the node
            configurations directly from CLI.
        custom_nodes_parent_subdir (:obj:`str` | :obj:`None`): Relative path to
            a folder which contains custom nodes that users have created to be
            used with PeekingDuck. For more information on using custom nodes,
            please refer to
            `Getting Started <getting_started/03_custom_nodes.html>`_.
        num_iter (int): Stop pipeline after running this number of iterations
        nodes (:obj:`List[AbstractNode]` | :obj:`None`): If a list of nodes is
            provided, initialize by the node stack directly.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        pipeline_path: Path = None,
        config_updates_cli: str = None,
        custom_nodes_parent_subdir: str = None,
        num_iter: int = None,
        nodes: List[AbstractNode] = None,
    ) -> None:
        self.logger = logging.getLogger(__name__)
        try:
            if nodes:
                # instantiated_nodes is created differently when given nodes
                self.pipeline = Pipeline(nodes)
            elif pipeline_path and config_updates_cli and custom_nodes_parent_subdir:
                # create Graph to run
                self.node_loader = DeclarativeLoader(
                    pipeline_path, config_updates_cli, custom_nodes_parent_subdir
                )
                self.pipeline = self.node_loader.get_pipeline()
            else:
                raise ValueError(
                    "Arguments error! Pass in either nodes to load directly via "
                    "Pipeline or pipeline_path, config_updates_cli, and "
                    "custom_nodes_parent_subdir to load via DeclarativeLoader."
                )
        except ValueError as error:
            self.logger.error(str(error))
            sys.exit(1)
        if RequirementChecker.n_update > 0:
            self.logger.warning(
                f"{RequirementChecker.n_update} package"
                f"{'s' * int(RequirementChecker.n_update > 1)} updated. "
                "Please rerun for the updates to take effect."
            )
            sys.exit(3)
        if num_iter is None or num_iter <= 0:
            self.num_iter = 0
        else:
            self.num_iter = num_iter
            self.logger.info(f"Run pipeline for {num_iter} iterations")

    def run(self) -> None:  # pylint: disable=too-many-branches
        """execute single or continuous inference"""
        num_iter = 0
        while not self.pipeline.terminate:
            for node in self.pipeline.nodes:
                if num_iter == 0:  # report node setup times at first iteration
                    self.logger.debug(f"First iteration: setup {node.name}...")
                    node_start_time = perf_counter()
                if self.pipeline.data.get("pipeline_end", False):
                    self.pipeline.terminate = True
                    if "pipeline_end" not in node.inputs:
                        continue

                if "all" in node.inputs:
                    inputs = copy.deepcopy(self.pipeline.data)
                else:
                    inputs = {
                        key: self.pipeline.data[key]
                        for key in node.inputs
                        if key in self.pipeline.data
                    }
                if hasattr(node, "optional_inputs"):
                    for key in node.optional_inputs:
                        # The nodes will not receive inputs with the optional
                        # key if it's not found upstream
                        if key in self.pipeline.data:
                            inputs[key] = self.pipeline.data[key]

                outputs = node.run(inputs)
                self.pipeline.data.update(outputs)
                if num_iter == 0:
                    node_end_time = perf_counter()
                    self.logger.debug(
                        f"{node.name} setup time = {node_end_time - node_start_time:.2f} sec"
                    )
            num_iter += 1
            if self.num_iter > 0 and num_iter >= self.num_iter:
                self.logger.info(f"Stopping pipeline after {num_iter} iterations")
                break

        # clean up nodes with threads
        for node in self.pipeline.nodes:
            if node.name.endswith(".visual"):
                node.release_resources()

    def get_pipeline(self) -> NodeList:
        """Retrieves run configuration.

        Returns:
            (:obj:`Dict`): Run configurations being used by runner.
        """
        return self.node_loader.node_list



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2020 YifuZhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Decodes model output to bbox and indices.

Modifications:
- Refactor mot_decode() to a class instead
- Remove unnecessary creation of lists since batch size 1 is hardcoded
- Hardcode ctdet_post_process to handle detections of batch size 1 since this
    assumptions is already made when calling the function
- Hardcode num_classes=1
    - Change post_process output type
    - Change ctdet_post_process output type
- Remove unnecessary .tolist() in ctdet_post_process
"""

from typing import Tuple

import numpy as np
import torch
import torch.nn.functional as F

from peekingduck.pipeline.nodes.model.fairmotv1.fairmot_files.utils import (
    gather_feat,
    transform_coords,
    transpose_and_gather_feat,
)


class Decoder:  # pylint: disable=too-few-public-methods
    """Decodes model output to bounding box coordinates and indices following
    the approach adopted by CenterNet.
    """

    def __init__(self, max_per_image: int, down_ratio: int) -> None:
        self.max_per_image = max_per_image
        self.down_ratio = down_ratio

    def __call__(  # pylint: disable=too-many-arguments
        self,
        heatmap: torch.Tensor,
        size: torch.Tensor,
        offset: torch.Tensor,
        orig_shape: Tuple[int, ...],
        input_shape: torch.Size,
    ) -> Tuple[np.ndarray, torch.Tensor]:
        """Decodes model outputs to bounding box coordinates and indices.

        Args:
            heatmap (torch.Tensor): A heatmap predicting where the object
                center will be.
            size (torch.Tensor): Size of the bounding boxes w.r.t. the object
                centers.
            offset (torch.Tensor): A continuous offset relative to the object
                centers to localize objects more precisely.
            orig_shape (Tuple[int, ...]): Shape of the original image.
            input_shape (torch.Size): Shape of the image fed to the model.

        Returns:
            (Tuple[np.ndarray, torch.Tensor]): A tuple containing detections
            and their respective indices. Indices are used to filter the Re-ID
            feature tensor.
        """
        batch, _, _, _ = heatmap.size()

        heatmap = self._nms(heatmap)

        scores, indices, classes, y_coords, x_coords = self._topk(heatmap)
        offset = transpose_and_gather_feat(offset, indices)
        offset = offset.view(batch, self.max_per_image, 2)
        x_coords = x_coords.view(batch, self.max_per_image, 1) + offset[:, :, 0:1]
        y_coords = y_coords.view(batch, self.max_per_image, 1) + offset[:, :, 1:2]
        size = transpose_and_gather_feat(size, indices)
        size = size.view(batch, self.max_per_image, 4)
        classes = classes.view(batch, self.max_per_image, 1)
        scores = scores.view(batch, self.max_per_image, 1)
        bboxes = torch.cat(
            [
                x_coords - size[..., 0:1],
                y_coords - size[..., 1:2],
                x_coords + size[..., 2:3],
                y_coords + size[..., 3:4],
            ],
            dim=2,
        )
        detections = torch.cat([bboxes, scores, classes], dim=2)
        detections = self._post_process(detections, orig_shape, input_shape)
        # Currently not needed as FairMOT only detect one class
        # detections = self._trim_outputs(detections)

        return detections, indices

    def _post_process(
        self,
        detections: torch.Tensor,
        orig_shape: Tuple[int, ...],
        input_shape: torch.Size,
    ) -> np.ndarray:
        """Post processes the detections following the approach by CenterNet.
        Translates/scales detections w.r.t. original image shape.

        Args:
            detections (torch.Tensor): Detections with the format
                [x1, y1, x2, y2, score, class] where (x1, y1) is top left and
                (x2, y2) is bottom right.
            orig_shape (Tuple[int, ...]): Shape of the original image.
            input_shape (torch.Size): Shape of the image fed to the model.

        Returns:
            (np.ndarray): Transformed detections w.r.t. the original image
            shape.
        """
        orig_h = float(orig_shape[0])
        orig_w = float(orig_shape[1])
        input_h = float(input_shape[2])
        input_w = float(input_shape[3])

        dets = detections.detach().cpu().numpy()
        dets = dets.reshape(1, -1, dets.shape[2])
        # Detection batch size has been hardcoded to 1 in FairMOT
        return _ctdet_post_process(
            dets[0].copy(),
            np.array([orig_w / 2.0, orig_h / 2.0], dtype=np.float32),
            max(input_w / input_h * orig_h, orig_w),
            (input_w // self.down_ratio, input_h // self.down_ratio),
        )

    def _topk(
        self, scores: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        """Selects top k scores and decodes to get xy coordinates.

        Args:
            scores (torch.Tensor): In the case of FairMOT, this is a heatmap
                predicting where the object center will be.

        Returns:
            (Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor,
            torch.Tensor]): Tuple containing top k detection scores and their
            respective indices, classes, y-, and x- coordinates.
        """
        k = self.max_per_image
        batch, cat, height, width = scores.size()

        topk_scores, topk_indices = torch.topk(scores.view(batch, cat, -1), k)

        topk_indices = (topk_indices % (height * width)).view(batch, -1, 1)
        topk_y_coords = torch.div(topk_indices, width, rounding_mode="floor")
        topk_x_coords = topk_indices % width

        topk_score, topk_index = torch.topk(topk_scores.view(batch, -1), k)
        topk_classes = torch.div(topk_index, k, rounding_mode="trunc")

        topk_indices = gather_feat(topk_indices, topk_index).view(batch, k)
        topk_y_coords = gather_feat(topk_y_coords, topk_index).view(batch, k)
        topk_x_coords = gather_feat(topk_x_coords, topk_index).view(batch, k)

        return topk_score, topk_indices, topk_classes, topk_y_coords, topk_x_coords

    def _trim_outputs(self, detections: np.ndarray) -> np.ndarray:  # pragma: no cover
        """In the case of multi-class detections, trims the output to be
        <=`self.max_per_image`.

        Args:
            detections (np.ndarray): Object detection results.

        Returns:
            (np.ndarray): Trimmed detection results.
        """
        if len(detections) > self.max_per_image:
            # FairMOT only detects one class so this is never called.
            kth = len(detections) - self.max_per_image
            cut_off = np.partition(detections[:, 4], kth)[kth]
            detections = detections[detections[:, 4] >= cut_off]
        return detections

    @staticmethod
    def _nms(heatmap: torch.Tensor, kernel: int = 3) -> torch.Tensor:
        """Uses maxpool to filter the max score and get local peaks.

        Args:
            heatmap (torch.Tensor): A heatmap predicting where the object
                center will be.
            kernel (int): Size of the window to take a max over.

        Returns:
            (torch.Tensor): Heatmap with only local peaks remaining.
        """
        pad = (kernel - 1) // 2

        hmax = F.max_pool2d(heatmap, (kernel, kernel), stride=1, padding=pad)
        keep = (hmax == heatmap).float()
        return heatmap * keep


def _ctdet_post_process(
    detections: np.ndarray,
    center: np.ndarray,
    scale: float,
    output_size: Tuple[float, float],
) -> np.ndarray:
    """Post-processes detections and translate/scale it back to the original
    image.

    Args:
        detections (np.ndarray): An array of detections each having the format
            [x1, y1, x2, y2, score, class] where (x1, y1) is top left and
            (x2, y2) is bottom right.
        center (np.ndarray): Coordinate of the center of the original image.
        scale (float): Scale between original image and input image fed to the
            model.
        output_size (Tuple[float, float]): Size of output by the model.

    Returns:
        (np.ndarray): Detections with coordinates transformed w.r.t. the
        original image.
    """
    detections[:, :2] = transform_coords(detections[:, :2], center, scale, output_size)
    detections[:, 2:4] = transform_coords(
        detections[:, 2:4], center, scale, output_size
    )
    classes = detections[:, -1]
    mask = classes == 0

    return np.concatenate([detections[mask, :4], detections[mask, 4:5]], axis=1).astype(
        np.float32
    )



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""MoveNet model with model types: singlepose lightning/thunder, multipose lightning"""

import logging
from typing import Any, Dict, Tuple

import numpy as np

from peekingduck.pipeline.nodes.model.movenetv1.movenet_files.predictor import Predictor
from peekingduck.weights_utils import checker, downloader, finder


class MoveNetModel:  # pylint: disable=too-few-public-methods
    """MoveNet model with model types: lightning, thunder"""

    def __init__(self, config: Dict[str, Any]) -> None:
        self.logger = logging.getLogger(__name__)

        # check threshold values
        if not 0 <= config["bbox_score_threshold"] <= 1:
            raise ValueError("bbox_score_threshold must be in [0, 1]")
        if not 0 <= config["keypoint_score_threshold"] <= 1:
            raise ValueError("keypoint_score_threshold must be in [0, 1]")
        if config["model_type"] not in [
            "singlepose_lightning",
            "singlepose_thunder",
            "multipose_lightning",
        ]:
            raise ValueError(
                "model_type must be one of ['singlepose_lightning', "
                "'singlepose_thunder', 'multipose_lightning']"
            )

        weights_dir, model_dir = finder.find_paths(
            config["root"], config["weights"], config["weights_parent_dir"]
        )

        # check for movenet weights, if none then download into weights folder
        if not checker.has_weights(weights_dir, model_dir):
            self.logger.info("---no weights detected. proceeding to download...---")
            downloader.download_weights(weights_dir, config["weights"]["blob_file"])
            self.logger.info(f"---weights downloaded to {weights_dir}.---")

        self.predictor = Predictor(config, model_dir)

    def predict(
        self, frame: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Predict poses from input frame

        Args:
            frame (np.array): image in numpy array

        Returns:
            Tuple of outputs of the model
            (bboxes, keypoints, keypoint_scores, keypoint_masks, keypoint_conns)

            bboxes (np.ndarray): Nx4 array of bboxes, N is number of detections
            keypoints (np.ndarray): Nx17x2 array of keypoint coordinates
            keypoints_scores (np.ndarray): Nx17 array of keypoint scores
            keypoints_conns (np.ndarray): NxD'x2 keypoint connections, where
                D' is the varying pair of valid keypoint connections per detection
        """
        return self.predictor.predict(frame)



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2020 YifuZhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Model components such as backbone, neck, and head for FairMOT.

Modifications include:
- Remove fill_fc_weights() and other fc weights initialization
- Remove fill_up_weights()
- Remove code branch when head_conv <= 0
- Avoid using loading model zoo weights to create fc layer in DLA
- DLASeg:
  - Remove base_name, pretrained, and out_channel arguments
  - Add default value for final_kernel, last_level, and head_conv
  - Hardcode final_kernel to 1
  - Remove load_base_weights
- DLA:
  - Use BasicBlock only
  - Remove linear_root argument
  - Omit creating num_classes member variable
  - Remove _make_level()
  - Remove dilation argument from _make_conv_level()
- IDAUp
  - Remove output_padding in Conv2DTranspose function call
"""

from typing import Dict, List

import numpy as np
import torch
from torch import nn

from peekingduck.pipeline.nodes.model.fairmotv1.fairmot_files.network_blocks import (
    BN_MOMENTUM,
    BasicBlock,
    DeformConv,
    Tree,
)


class DLASeg(nn.Module):
    """The encoder-decoder network comprising of:
    - ResNet-34 with an enhanced Deep Layer Aggregation (DLA) as backbone
    - CenterNet as the detection branch
    - FairMOT EmbeddingHead as the Re-ID branch

    Args:
        heads (Dict[str, int]): Configuration for the output channels for the
            various heads of the model.
        down_ratio (int): The downscale ratio from images to heatmap. In the
            case of FairMOT, this is 4.
        last_level (int): The last level of input feature fed into the
            upsampling block. Default is 5.
        head_conv (int): Number of channels in all heads. Default is 256.
    """

    # pylint: disable=redefined-builtin

    def __init__(
        self,
        heads: Dict[str, int],
        down_ratio: int,
        last_level: int = 5,
        head_conv: int = 256,
    ) -> None:
        super().__init__()
        self.first_level = int(np.log2(down_ratio))
        self.last_level = last_level
        self.base = DLA([1, 1, 1, 2, 2, 1], [16, 32, 64, 128, 256, 512])
        channels = self.base.channels
        scales = [2 ** i for i in range(len(channels[self.first_level :]))]
        self.dla_up = DLAUp(self.first_level, channels[self.first_level :], scales)

        self.ida_up = IDAUp(
            channels[self.first_level : self.last_level],
            channels[self.first_level],
            [2 ** i for i in range(self.last_level - self.first_level)],
        )

        self.heads = heads
        for head in self.heads:
            classes = self.heads[head]
            head_network = nn.Sequential(
                nn.Conv2d(
                    channels[self.first_level],
                    head_conv,
                    kernel_size=3,
                    padding=1,
                    bias=True,
                ),
                nn.ReLU(inplace=True),
                nn.Conv2d(
                    head_conv, classes, kernel_size=1, stride=1, padding=0, bias=True
                ),
            )
            # Sets the dict key as the member variable
            self.__setattr__(head, head_network)

    def forward(  # pylint: disable=invalid-name
        self, input: torch.Tensor
    ) -> Dict[str, torch.Tensor]:
        """Defines the computation performed at every call.

        Args:
            input (torch.Tensor): Input from the previous layer.

        Returns:
            (Dict[str, torch.Tensor]): A dictionary of tensors with keys
            corresponding to `self.heads`.
        """
        layers = self.base(input)
        layers = self.dla_up(layers)

        y = [layers[i].clone() for i in range(self.last_level - self.first_level)]
        self.ida_up(y, 0, len(y))

        outputs = {}
        for head in self.heads:
            outputs[head] = getattr(self, head)(y[-1])
        return outputs


class DLA(nn.Module):  # pylint: disable=too-many-instance-attributes
    """Deep Layer Aggregation to be used as the backbone of FairMOT's
    encoder-decoder network.

    Args:
        levels (List[int]): List of aggregation depths at various stages.
        channels (List[int]): List of number of channels at various stages.
        num_classes (int): Number of classes for classification. NOTE: Not used
            in FairMOT is needed to properly load the model weights.
        residual_root (bool): Flag to indicate if a residual layer should be
            used in the root block. Default is False.
    """

    # pylint: disable=redefined-builtin

    def __init__(
        self,
        levels: List[int],
        channels: List[int],
        num_classes: int = 1000,
    ) -> None:
        super().__init__()
        # DLA-34 (used by FairMOT) uses basic blocks as the residual block
        block = BasicBlock
        self.channels = channels
        self.base_layer = nn.Sequential(
            nn.Conv2d(3, channels[0], kernel_size=7, stride=1, padding=3, bias=False),
            nn.BatchNorm2d(channels[0], momentum=BN_MOMENTUM),
            nn.ReLU(inplace=True),
        )
        self.level0 = self._make_conv_level(channels[0], channels[0], levels[0])
        self.level1 = self._make_conv_level(
            channels[0], channels[1], levels[1], stride=2
        )
        self.level2 = Tree(
            levels[2],
            block,
            channels[1],
            channels[2],
            2,
            level_root=False,
        )
        self.level3 = Tree(
            levels[3],
            block,
            channels[2],
            channels[3],
            2,
            level_root=True,
        )
        self.level4 = Tree(
            levels[4],
            block,
            channels[3],
            channels[4],
            2,
            level_root=True,
        )
        self.level5 = Tree(
            levels[5],
            block,
            channels[4],
            channels[5],
            2,
            level_root=True,
        )
        # Apparently not needed
        self.fc = nn.Conv2d(  # pylint: disable=invalid-name
            self.channels[-1],
            num_classes,
            kernel_size=1,
            stride=1,
            padding=0,
            bias=True,
        )

    def forward(self, input: torch.Tensor) -> List[torch.Tensor]:
        """Defines the computation performed at every call.

        Args:
            input (torch.Tensor): Input from the previous layer.

        Returns:
            (List[torch.Tensor]): A list of tensors containing the output at
            every stage.
        """
        outputs = []
        input = self.base_layer(input)
        for i in range(6):
            input = getattr(self, f"level{i}")(input)
            outputs.append(input)
        return outputs

    @staticmethod
    def _make_conv_level(
        in_channels: int, out_channels: int, num_convs: int, stride: int = 1
    ) -> nn.Sequential:
        """Creates the networks for one of the earlier stages in DLA.

        Args:
            in_channels (int): Number of channels in the input image.
            out_channels (int): Number of channels producted by the
                convolution.
            num_convs (int): Number of convolution layers in the stage.
            stride (int): Stride of the convolution. Default is 1.

        Returns:
            (nn.Sequential): A sequential container of all the networks in the
            stage.
        """
        modules = []
        for i in range(num_convs):
            modules.extend(
                [
                    nn.Conv2d(
                        in_channels,
                        out_channels,
                        kernel_size=3,
                        stride=stride if i == 0 else 1,
                        padding=1,
                        bias=False,
                    ),
                    nn.BatchNorm2d(out_channels, momentum=BN_MOMENTUM),
                    nn.ReLU(inplace=True),
                ]
            )
            in_channels = out_channels
        return nn.Sequential(*modules)


class DLAUp(nn.Module):
    """DLA upsample network.

    Args:
        start_level (int): The starting stage of this upsample network.
        channels (List[int]): List of number of channels at various stages.
        scales (List[int]): Scale factors at various stages.
        in_channels (List[int]): Number of channels in the input image at
            various stages.
    """

    def __init__(
        self,
        start_level: int,
        channels: List[int],
        scales: List[int],
        in_channels: List[int] = None,
    ) -> None:
        super().__init__()
        self.start_level = start_level
        if in_channels is None:
            in_channels = channels
        self.channels = channels
        scales_arr = np.array(scales, dtype=int)
        for i in range(len(channels) - 1):
            j = -i - 2
            setattr(
                self,
                f"ida_{i}",
                IDAUp(in_channels[j:], channels[j], scales_arr[j:] // scales_arr[j]),
            )
            scales_arr[j + 1 :] = scales_arr[j]
            in_channels[j + 1 :] = [channels[j] for _ in channels[j + 1 :]]

    def forward(self, layers: List[torch.Tensor]) -> List[torch.Tensor]:
        """Defines the computation performed at every call.

        Args:
            layers (List[torch.Tensor]): Inputs from the various stages.

        Returns:
            (List[torch.Tensor]): A list of tensors containing the output at
            every stage.
        """
        out = [layers[-1]]  # start with 32
        for i in range(len(layers) - self.start_level - 1):
            ida = getattr(self, f"ida_{i}")
            ida(layers, len(layers) - i - 2, len(layers))
            out.insert(0, layers[-1])
        return out


class IDAUp(nn.Module):
    """Iterative Deep Aggregation network.

    Args:
        in_channels_list (List[int]): List of Number of channels in the input
            image at various stages.
        out_channels (int): Number of channels producted by the convolution.
        up_strides (List[int]): List of strides for upsampling at various
            stages.
    """

    def __init__(
        self, in_channels_list: List[int], out_channels: int, up_strides: List[int]
    ) -> None:
        super().__init__()
        for i in range(1, len(in_channels_list)):
            in_channels = in_channels_list[i]
            stride = int(up_strides[i])

            project = DeformConv(in_channels, out_channels)
            node = DeformConv(out_channels, out_channels)
            upsample = nn.ConvTranspose2d(
                out_channels,
                out_channels,
                stride * 2,
                stride=stride,
                padding=stride // 2,
                groups=out_channels,
                bias=False,
            )

            setattr(self, f"proj_{i}", project)
            setattr(self, f"up_{i}", upsample)
            setattr(self, f"node_{i}", node)

    def forward(
        self, layers: List[torch.Tensor], start_level: int, end_level: int
    ) -> None:
        """Defines the computation performed at every call. NOTE: This modifies
        ``layers`` in-place.

        Args:
            layers (List[torch.Tensor]): Inputs from the various stages.
            start_level (int): The starting stage number.
            end_level (int): The ending stage number.
        """
        for i in range(start_level + 1, end_level):
            project = getattr(self, f"proj_{i - start_level}")
            node = getattr(self, f"node_{i - start_level}")
            upsample = getattr(self, f"up_{i - start_level}")

            layers[i] = upsample(project(layers[i]))
            layers[i] = node(layers[i] + layers[i - 1])



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright 2021 Megvii, Base Detection
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A collection of utility functions used by YOLOX.

Modifications include:
- Refactor fuse_conv_and_bn and fuse_model
- Adding xywh2xyxy and xyxy2xyxyn
"""

from pathlib import Path
from typing import no_type_check

import torch
import torch.nn as nn

from peekingduck.pipeline.nodes.model.yoloxv1.yolox_files.model import YOLOX
from peekingduck.pipeline.nodes.model.yoloxv1.yolox_files.network_blocks import BaseConv


@no_type_check
def fuse_conv_and_bn(conv: nn.Conv2d, batch_norm: nn.BatchNorm2d) -> nn.Conv2d:
    """Fuses convolution and batchnorm layers.

    Reference: https://tehnokv.com/posts/fusing-batchnorm-and-conv/
    """
    fusedconv = nn.Conv2d(
        conv.in_channels,
        conv.out_channels,
        conv.kernel_size,
        conv.stride,
        conv.padding,
        groups=conv.groups,
        bias=True,
        device=conv.weight.device,
        dtype=conv.weight.dtype,
    ).requires_grad_(False)

    # prepare filters
    w_conv = conv.weight.clone().view(conv.out_channels, -1)
    w_bn = torch.diag(
        batch_norm.weight.div(torch.sqrt(batch_norm.eps + batch_norm.running_var))
    )
    fusedconv.weight.copy_(torch.mm(w_bn, w_conv).view(fusedconv.weight.shape))

    # prepare spatial bias
    b_conv = (
        torch.zeros(
            conv.weight.size(0), dtype=conv.weight.dtype, device=conv.weight.device
        )
        if conv.bias is None
        else conv.bias
    )
    b_bn = batch_norm.bias - batch_norm.weight.mul(batch_norm.running_mean).div(
        torch.sqrt(batch_norm.running_var + batch_norm.eps)
    )
    fusedconv.bias.copy_(torch.mm(w_bn, b_conv.reshape(-1, 1)).reshape(-1) + b_bn)

    return fusedconv


def fuse_model(model: YOLOX) -> YOLOX:
    """Fuses the batch normalization layers in `BaseConv` modules."""
    for module in model.modules():
        if isinstance(module, BaseConv) and hasattr(module, "bn"):
            module.conv = fuse_conv_and_bn(module.conv, module.bn)  # update conv
            delattr(module, "bn")  # remove batchnorm
            # update forward
            module.forward = module.fuseforward  # type: ignore
    return model


def strip_optimizer(
    weights_path: Path, key_to_keep: str = "model", half: bool = False
) -> None:  # pragma: no cover
    """Removes non-model items from the weights file.

    The inference Detector requires only "model" parameters from the weights
    file. Deletes all other values from the weights file. Optionally, converts
    all float and double parameters to half-precision. Leaves int parameters
    such as `num_batches_tracked` in `BatchNorm2d` untouched.

    Args:
        weights_path (Path): Path to weights file.
        key_to_keep (str): The key which contains the models weights. Default
            is "model".
        half (bool): Flag to determine if float and double parameters should
            be converted to half-precision.
    """
    stripped_weights_path = weights_path.with_name(
        f"{weights_path.stem}-stripped"
        f"{'-half' if half else ''}{weights_path.suffix}"
    )
    orig_filesize = weights_path.stat().st_size / 1e6
    ckpt = torch.load(str(weights_path), map_location=torch.device("cpu"))
    # Remove all data other than "model", such as amp, optimizer, start_epoch
    delete_keys = [key for key in ckpt.keys() if key != key_to_keep]
    for key in delete_keys:
        del ckpt[key]
    for param in ckpt[key_to_keep]:
        # Only convert double and float to half-precision
        if half and ckpt[key_to_keep][param].dtype in (torch.double, torch.float):
            ckpt[key_to_keep][param] = ckpt[key_to_keep][param].half()
        ckpt[key_to_keep][param].requires_grad = False

    torch.save(ckpt, str(stripped_weights_path))
    stripped_filesize = stripped_weights_path.stat().st_size / 1e6
    print(
        f"Saved as {stripped_weights_path}. "
        f"Original size: {orig_filesize}MB. "
        f"Stripped size: {stripped_filesize}MB."
    )



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility functions to load the MTCNN model
"""

import logging
from typing import Tuple, Callable

import tensorflow as tf

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


def wrap_frozen_graph(graph_def: tf.compat.v1.GraphDef) -> Callable:
    """
    Wraps a frozen graph into a function

    Args:
        graph_def (tf.compat.v1.GraphDef): A frozen graph in graph_def format

    Return:
        wrapped_import (tensorflow.python.eager.wrap_function.WrappedFunction):
        A wrapped_import function to perform your inference with
    """

    def _imports_graph_def(
        img: tf.Tensor, min_size: tf.Tensor, factor: tf.Tensor, thresholds: tf.Tensor
    ) -> Tuple[tf.Tensor, tf.Tensor, tf.Tensor]:
        prob, landmarks, box = tf.compat.v1.import_graph_def(
            graph_def,
            input_map={
                "input:0": img,
                "min_size:0": min_size,
                "thresholds:0": thresholds,
                "factor:0": factor,
            },
            return_elements=["prob:0", "landmarks:0", "box:0"],
            name="",
        )

        return box, prob, landmarks

    wrapped_import = tf.compat.v1.wrap_function(
        _imports_graph_def,
        [
            tf.TensorSpec(shape=[None, None, 3], dtype=tf.float32),
            tf.TensorSpec(shape=[], dtype=tf.float32),
            tf.TensorSpec(shape=[], dtype=tf.float32),
            tf.TensorSpec(shape=[3], dtype=tf.float32),
        ],
    )

    return wrapped_import


def load_graph(filename: str) -> tf.function:
    """
    Loads a frozen graph and wraps it into a function

    Args:
        filename (str): Path to the frozen graph or model

    Return:
        wrapped_import (tensorflow.python.eager.wrap_function.WrappedFunction):
        A wrapped_import function to perform your inference with
    """
    with tf.io.gfile.GFile(filename, "rb") as graph_file:
        graph_def = tf.compat.v1.GraphDef()
        graph_def.ParseFromString(graph_file.read())

        frozen_func = wrap_frozen_graph(graph_def=graph_def)

        return frozen_func



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Estimates the 3D coordinates of a person given 2D pose coordinates.
"""

from typing import Any, Dict

import numpy as np

from peekingduck.pipeline.nodes.node import AbstractNode

NOSE = 0
LEFT_SHOULDER = 5
RIGHT_SHOULDER = 6
LEFT_PELVIS = 11
RIGHT_PELVIS = 12
TORSO_KEYPOINTS = [NOSE, LEFT_SHOULDER, RIGHT_SHOULDER, LEFT_PELVIS, RIGHT_PELVIS]


class Node(AbstractNode):
    """Uses pose keypoint information of the torso to estimate 3D location.

    Inputs:
        |keypoints_data|

    Outputs:
        |obj_3D_locs_data|

    Configs:
        focal_length (:obj:`float`): **default = 1.14**. |br|
            Approximate focal length of webcam used, in metres. Example on
            measuring focal length can be found `here <https://learnopencv.com
            /approximate-focal-length-for-webcams-and-cell-phone-cameras/>`_.
        torso_factor (:obj:`float`): **default = 0.9**. |br|
            A factor used to estimate real-world distance from pixels, based on
            average human torso length in metres. The value varies across
            different camera set-ups, and calibration may be required.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Converts pose keypoints into 3D locations."""
        locations = []

        for keypoints in inputs["keypoints"]:
            torso_keypoints = self._get_torso_keypoints(keypoints)
            if self._enough_torso_keypoints(torso_keypoints):
                bbox = self._get_bbox(torso_keypoints)
            else:
                bbox = self._get_bbox(keypoints)

            point = self._get_3d_point_from_bbox(
                bbox, self.focal_length, self.torso_factor
            )
            locations.append(point)

        outputs = {"obj_3D_locs": locations}

        return outputs

    @staticmethod
    def _get_torso_keypoints(keypoints: np.ndarray) -> np.ndarray:
        """Filter keypoints to get only selected keypoints for torso"""
        torso_keypoints = keypoints[TORSO_KEYPOINTS, :]  # type: ignore
        # ignore keypoints that are '-1.' as below confidence score and are masked
        torso_keypoints = np.reshape(torso_keypoints[torso_keypoints != -1.0], (-1, 2))

        return torso_keypoints

    @staticmethod
    def _enough_torso_keypoints(torso_keypoints: np.ndarray) -> bool:
        """Returns False if not enough keypoints to represent torso"""
        if torso_keypoints.shape[0] >= 2:
            return True
        return False

    @staticmethod
    def _get_bbox(keypoints: np.ndarray) -> np.ndarray:
        """Get coordinates of a bbox around keypoints"""
        top_left_x, top_left_y = keypoints.min(axis=0)
        btm_right_x, btm_right_y = keypoints.max(axis=0)

        return np.array([top_left_x, top_left_y, btm_right_x, btm_right_y])

    @staticmethod
    def _get_3d_point_from_bbox(
        bbox: np.ndarray, focal_length: float, torso_factor: float
    ) -> np.ndarray:
        """Get the 3d coordinates of the centre of a bounding box"""
        # Subtraction is to make the camera the origin of the coordinate system
        center_2d = ((bbox[0:2] + bbox[2:4]) * 0.5) - np.array([0.5, 0.5])
        torso_height = bbox[3] - bbox[1]

        z_coord = (focal_length * torso_factor) / torso_height
        x_coord = (center_2d[0] * torso_factor) / torso_height
        y_coord = (center_2d[1] * torso_factor) / torso_height

        return np.array([x_coord, y_coord, z_coord])



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright 2021 Megvii, Base Detection
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""YOLOX model with its backbone (YOLOFAPN) and head (YOLOXHead).

Modifications include:
- YOLOX
    - Uses only YOLOPAFPN as backbone
    - Uses only YOLOHead as head
- YOLOPAFPN
    - Refactor arguments name
- YOLOXHead
    - Uses range based loop to iterate through in_channels
    - Removed training-related code and arguments
        - Code under the `if self.training` scope
        - get_output_and_grid() and initialize_biases() methods
"""

from typing import Any, List, Tuple

import torch
import torch.nn as nn

from peekingduck.pipeline.nodes.model.yoloxv1.yolox_files.darknet import CSPDarknet
from peekingduck.pipeline.nodes.model.yoloxv1.yolox_files.network_blocks import (
    BaseConv,
    CSPLayer,
)

IN_CHANNELS = [256, 512, 1024]


class YOLOX(nn.Module):
    """YOLOX model module.

    The module list is defined by create_yolov3_modules function. The network
    returns loss values from three YOLO layers during training and detection
    results during test.
    """

    # pylint: disable=arguments-differ
    def __init__(
        self,
        num_classes: int,
        depth: float,
        width: float,
    ) -> None:
        super().__init__()
        self.backbone = YOLOPAFPN(depth, width)
        self.head = YOLOXHead(num_classes, width)
        self.apply(YOLOX.initialize_batch_norm)

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        """Defines the computation performed at every call.

        Args:
            inputs (torch.Tensor): Input image.

        Returns:
            (torch.Tensor): The decoded output with the shape (B,D,85) where
            B is the batch size, D is the number of detections. The 85 columns
            consist of the following values:
            [x, y, w, h, conf, (cls_conf of the 80 COCO classes)].
        """
        # FPN output content features of [dark3, dark4, dark5]
        fpn_outs = self.backbone(inputs)
        return self.head(fpn_outs)

    @staticmethod
    def initialize_batch_norm(module: nn.Module) -> None:
        """Initializes the BatchNorm2d layers."""
        for mod in module.modules():
            if isinstance(mod, nn.BatchNorm2d):
                mod.eps = 1e-3
                mod.momentum = 0.03


class YOLOPAFPN(nn.Module):  # pylint: disable=too-many-instance-attributes
    """YOLOv3 model. Darknet 53 is the default backbone of this model."""

    # pylint: disable=arguments-differ, dangerous-default-value, invalid-name
    def __init__(
        self,
        depth: float = 1.0,
        width: float = 1.0,
    ) -> None:
        super().__init__()
        n_bottleneck = round(3 * depth)
        self.in_features = ("dark3", "dark4", "dark5")
        self.backbone = CSPDarknet(depth, width, self.in_features)

        self.upsample = nn.Upsample(scale_factor=2, mode="nearest")
        N256, N512, N1024 = IN_CHANNELS
        self.lateral_conv0 = BaseConv(int(N1024 * width), int(N512 * width), 1, 1)
        self.C3_p4 = self.make_csp_layer(N512, N512, n_bottleneck, width)

        self.reduce_conv1 = BaseConv(int(N512 * width), int(N256 * width), 1, 1)
        self.C3_p3 = self.make_csp_layer(N256, N256, n_bottleneck, width)

        # bottom-up conv
        self.bu_conv2 = BaseConv(int(N256 * width), int(N256 * width), 3, 2)
        self.C3_n3 = self.make_csp_layer(N256, N512, n_bottleneck, width)

        # bottom-up conv
        self.bu_conv1 = BaseConv(int(N512 * width), int(N512 * width), 3, 2)
        self.C3_n4 = self.make_csp_layer(N512, N1024, n_bottleneck, width)

    @staticmethod
    def make_csp_layer(
        in_channel: int, out_channel: int, depth: int, width: float
    ) -> CSPLayer:
        """Returns a CSPLayer.

        Args:
            in_channel (int): Input channels.
            out_channel (int): Output channels.
            depth (int): Number of Bottlenecks.
            width (float): Multiplier to scale the number of input and output
                channels.

        Returns:
            (CSPLayer): A CSPLayer consisting of following blocks:
            Conv -> Bottlenecks -> Conv
                               /
                        Conv -
        """
        return CSPLayer(
            int(2 * in_channel * width), int(out_channel * width), depth, False
        )

    def forward(
        self, inputs: torch.Tensor
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """Defines the computation performed at every call.

        Args:
            inputs: Input images.

        Returns:
            (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): FPN feature.
        """
        #  backbone
        out_features = self.backbone(inputs)
        [x2, x1, x0] = [out_features[f] for f in self.in_features]

        fpn_out0 = self.lateral_conv0(x0)  # 1024->512/32
        f_out0 = self.upsample(fpn_out0)  # 512/16
        f_out0 = torch.cat([f_out0, x1], 1)  # 512->1024/16
        f_out0 = self.C3_p4(f_out0)  # 1024->512/16

        fpn_out1 = self.reduce_conv1(f_out0)  # 512->256/16
        f_out1 = self.upsample(fpn_out1)  # 256/8
        f_out1 = torch.cat([f_out1, x2], 1)  # 256->512/8
        pan_out2 = self.C3_p3(f_out1)  # 512->256/8

        p_out1 = self.bu_conv2(pan_out2)  # 256->256/16
        p_out1 = torch.cat([p_out1, fpn_out1], 1)  # 256->512/16
        pan_out1 = self.C3_n3(p_out1)  # 512->512/16

        p_out0 = self.bu_conv1(pan_out1)  # 512->512/32
        p_out0 = torch.cat([p_out0, fpn_out0], 1)  # 512->1024/32
        pan_out0 = self.C3_n4(p_out0)  # 1024->1024/32

        return pan_out2, pan_out1, pan_out0


class YOLOXHead(nn.Module):  # pylint: disable=too-many-instance-attributes
    """Decoupled head.

    The decoupled head contains two parallel branches for classification and
    regression tasks. An IoU branch is added to the regression branch after
    the conv layers.
    """

    # pylint: disable=arguments-differ, dangerous-default-value, invalid-name
    def __init__(
        self,
        num_classes: int,
        width: float = 1.0,
        strides: List[int] = [8, 16, 32],
    ) -> None:
        super().__init__()

        self.sizes: List[Tuple[Any, ...]]
        feat_channels = int(256 * width)
        self.n_anchors = 1
        self.num_classes = num_classes

        self.cls_convs = nn.ModuleList()
        self.reg_convs = nn.ModuleList()
        self.cls_preds = nn.ModuleList()
        self.reg_preds = nn.ModuleList()
        self.obj_preds = nn.ModuleList()
        self.stems = nn.ModuleList()

        for in_channel in IN_CHANNELS:
            self.stems.append(BaseConv(int(in_channel * width), feat_channels, 1, 1))
            self.cls_convs.append(nn.Sequential(*self.make_group_layer(feat_channels)))
            self.reg_convs.append(nn.Sequential(*self.make_group_layer(feat_channels)))
            self.cls_preds.append(
                nn.Conv2d(feat_channels, self.n_anchors * self.num_classes, 1, 1, 0)
            )
            self.reg_preds.append(nn.Conv2d(feat_channels, 4, 1, 1, 0))
            self.obj_preds.append(nn.Conv2d(feat_channels, self.n_anchors * 1, 1, 1, 0))

        self.strides = strides
        self.grids = [torch.zeros(1)] * len(IN_CHANNELS)

    @staticmethod
    def make_group_layer(in_channels: int) -> Tuple[BaseConv, BaseConv]:
        """2x BaseConv layer.

        Args:
            in_channels (int): The number of input and output channels for
                BaseConv.

        Returns:
            (Tuple[BaseConv, BaseConv]): A tuple containing 2 BaseConv blocks.
        """
        return (
            BaseConv(in_channels, in_channels, 3, 1),
            BaseConv(in_channels, in_channels, 3, 1),
        )

    def forward(
        self, xin: Tuple[torch.Tensor, torch.Tensor, torch.Tensor]
    ) -> torch.Tensor:
        """Defines the computation performed at every call.

        Args:
            xin (Tuple[torch.Tensor, torch.Tensor, torch.Tensor]): Inputs from
                `YOLOPAFPN`, contains 3 levels of FPN features (256, 512,
                and 1024).

        Returns:
            (torch.Tensor): The decoded output with the shape (B,D,85) where
            B is the batch size, D is the number of detections. The 85 columns
            consist of the following values:
            [x, y, w, h, conf, (cls_conf of the 80 COCO classes)].
        """
        outputs = []
        for k, (cls_conv, reg_conv, x) in enumerate(
            zip(self.cls_convs, self.reg_convs, xin)
        ):
            x = self.stems[k](x)

            cls_feat = cls_conv(x)
            cls_output = self.cls_preds[k](cls_feat)

            reg_feat = reg_conv(x)
            reg_output = self.reg_preds[k](reg_feat)
            obj_output = self.obj_preds[k](reg_feat)

            output = torch.cat(
                [reg_output, obj_output.sigmoid(), cls_output.sigmoid()], 1
            )
            outputs.append(output)

        self.sizes = [x.shape[-2:] for x in outputs]
        # [batch, n_anchors_all, 85]
        outputs_tensor = torch.cat(
            [x.flatten(start_dim=2) for x in outputs], dim=2
        ).permute(0, 2, 1)
        # Always decode output for inference
        outputs_tensor = self.decode_outputs(outputs_tensor, xin[0].type())
        return outputs_tensor

    def decode_outputs(self, outputs: torch.Tensor, dtype: str) -> torch.Tensor:
        """Converts raw output to [x, y, w, h] format.

        Args:
            outputs (torch.Tensor): Raw output tensor. The first 4 columns
                contain 2 offsets in terms of the top-left corner of the grid,
                and the height and width of the predicted box. The rest of the
                columns are not accessed in this method.
            dtype (str): Data type.

        Returns:
            (torch.Tensor): The decoded output with the shape (B,D,85) where
            B is the batch size, D is the number of detections. The 85 columns
            consist of the following values:
            [x, y, w, h, conf, (cls_conf of the 80 COCO classes)].
        """
        grids = []
        strides = []
        for (hsize, wsize), stride in zip(self.sizes, self.strides):
            yv, xv = torch.meshgrid([torch.arange(hsize), torch.arange(wsize)])
            grid = torch.stack((xv, yv), 2).view(1, -1, 2)
            grids.append(grid)
            shape = grid.shape[:2]
            strides.append(torch.full((*shape, 1), stride))

        grids_tensor = torch.cat(grids, dim=1).type(dtype)
        strides_tensor = torch.cat(strides, dim=1).type(dtype)

        outputs[..., :2] = (outputs[..., :2] + grids_tensor) * strides_tensor
        outputs[..., 2:4] = torch.exp(outputs[..., 2:4]) * strides_tensor
        return outputs



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2020 YifuZhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""FairMOT model for human detection and tracking."""

import logging
from typing import Any, Dict, List, Tuple, Union

import numpy as np

from peekingduck.pipeline.nodes.model.fairmotv1.fairmot_files.tracker import Tracker
from peekingduck.weights_utils import checker, downloader, finder


class FairMOTModel:  # pylint: disable=too-few-public-methods
    """FairMOT Model for person tracking.
    Args:
        config (Dict[str, Any]): Model configuration options.
        frame_rate (float): The frame rate of the current video sequence,
            used for computing the size of track buffer.
    Raises:
        ValueError: `score_threshold` is beyond [0, 1].
        ValueError: `K` is less than 1.
        ValueError: `min_box_area` is less than 1.
        ValueError: `track_buffer` is less than 1.
    """

    def __init__(self, config: Dict[str, Any], frame_rate: float) -> None:
        self.logger = logging.getLogger(__name__)

        # Check threshold values
        if not 0 < config["score_threshold"] <= 1:
            raise ValueError("score_threshold must be in [0, 1]")
        ensure_more_than_zero(config, ["K", "min_box_area", "track_buffer"])

        # Check for weights
        weights_dir, model_dir = finder.find_paths(
            config["root"], config["weights"], config["weights_parent_dir"]
        )
        if not checker.has_weights(weights_dir, model_dir):
            self.logger.info("No weights detected. Proceeding to download...")
            downloader.download_weights(weights_dir, config["weights"]["blob_file"])
            self.logger.info(f"Weights downloaded to {weights_dir}.")

        self.tracker = Tracker(config, model_dir, frame_rate)

    def predict(
        self, image: np.ndarray
    ) -> Tuple[List[np.ndarray], List[str], List[float], List[int]]:
        """Track objects from image.

        Args:
            image (np.ndarray): Image in numpy array.

        Returns:
            (Tuple[List[np.ndarray], List[str], List[float], List[int]]): A tuple of
            - Numpy array of detected bounding boxes.
            - List of detection class labels (person).
            - List of detection confidence scores.
            - List of track IDs.

        Raises:
            TypeError: The provided `image` is not a numpy array.
        """
        if not isinstance(image, np.ndarray):
            raise TypeError("image must be a np.ndarray")
        bboxes, track_ids, bbox_scores = self.tracker.track_objects_from_image(image)
        bbox_labels = ["person"] * len(bboxes)
        return bboxes, bbox_labels, bbox_scores, track_ids


def ensure_more_than_zero(config: Dict[str, Any], key: Union[List[str], str]) -> None:
    """Checks that configuration values specified by ``key`` is more than zero.

    Args:
        config (Dict[str, Any]): Dictionary containing various configuration
            values.
        key (Union[List[str], str]): List of configuration keys to check for.

    Raises:
        TypeError: ``key`` is not in (List[str], str).
        ValueError: If the configuration value is <=0.
    """
    if isinstance(key, str):
        if config[key] < 1:
            raise ValueError(f"{key} must be more than 0")
    elif isinstance(key, list):
        for k in key:
            ensure_more_than_zero(config, k)
    else:
        raise TypeError("'key' must be either 'str' or 'list'")



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2020 YifuZhang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Modifications:
- Rename planes to channels to be consistent to Conv2d argument naming
- Rearrange argument order in Tree
- Remove root_residual argument in Tree
- Remove self.residual in Root
"""

from typing import Callable, List, Union

import torch
from torch import nn

from peekingduck.pipeline.nodes.model.fairmotv1.fairmot_files.dcn import DCNv2

BN_MOMENTUM = 0.1


class BasicBlock(nn.Module):
    """Basic residual block structure used by DLA.

    Args:
        in_channels (int): Number of channels in the input image.
        out_channels (int): Number of channels produced by the block.
        stride (int): Stride of the convolution. Default is 1.
        dilation (int): Spacing between kernel elements. Default is 1.
    """

    # pylint: disable=redefined-builtin

    def __init__(
        self, in_channels: int, out_channels: int, stride: int = 1, dilation: int = 1
    ) -> None:
        super().__init__()
        self.conv1 = nn.Conv2d(
            in_channels,
            out_channels,
            kernel_size=3,
            stride=stride,
            padding=dilation,
            bias=False,
            dilation=dilation,
        )
        self.bn1 = nn.BatchNorm2d(out_channels, momentum=BN_MOMENTUM)
        self.relu = nn.ReLU(inplace=True)
        self.conv2 = nn.Conv2d(
            out_channels,
            out_channels,
            kernel_size=3,
            stride=1,
            padding=dilation,
            bias=False,
            dilation=dilation,
        )
        self.bn2 = nn.BatchNorm2d(out_channels, momentum=BN_MOMENTUM)
        self.stride = stride

    def forward(
        self, input: torch.Tensor, residual: torch.Tensor = None
    ) -> torch.Tensor:
        """Defines the computation performed at every call.

        Args:
            input (torch.Tensor): Input from the previous layer.

        Returns:
            (torch.Tensor): The output tensor of the block.
        """
        if residual is None:
            residual = input

        out = self.conv1(input)
        out = self.bn1(out)
        out = self.relu(out)

        out = self.conv2(out)
        out = self.bn2(out)

        out += residual
        out = self.relu(out)

        return out


class DeformConv(nn.Module):
    """Performs deformable convolution followed by batch norm.

    Args:
        in_channels (int): Number of channels in the input image.
        out_channels (int): Number of channels produced by the convolution.
    """

    # pylint: disable=redefined-builtin

    def __init__(self, in_channels: int, out_channels: int) -> None:
        super().__init__()
        self.actf = nn.Sequential(
            nn.BatchNorm2d(out_channels, momentum=BN_MOMENTUM), nn.ReLU(inplace=True)
        )
        self.conv = DCNv2(
            in_channels,
            out_channels,
            kernel_size=(3, 3),
            stride=1,
            padding=1,
            dilation=1,
            deformable_groups=1,
        )

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        """Defines the computation performed at every call.

        Args:
            input (torch.Tensor): Input from the previous layer.

        Returns:
            (torch.Tensor): The output tensor of the block.
        """
        out = self.conv(input)
        out = self.actf(out)
        return out


class Root(nn.Module):
    """Root node in Hierarchical Deep Aggregation.

    Args:
        in_channels (int): Number of channels in the input image.
        out_channels (int): Number of channels produced by the block.
        kernel_size (int): Size of the convolving kernel, used for calculating
            padding size only.
        residual (bool): Flag to indicate if a residual layer should be used.
    """

    def __init__(self, in_channels: int, out_channels: int, kernel_size: int) -> None:
        super().__init__()
        self.conv = nn.Conv2d(
            in_channels,
            out_channels,
            1,
            stride=1,
            bias=False,
            padding=(kernel_size - 1) // 2,
        )
        self.bn = nn.BatchNorm2d(  # pylint: disable=invalid-name
            out_channels, momentum=BN_MOMENTUM
        )
        self.relu = nn.ReLU(inplace=True)

    def forward(self, inputs: List[torch.Tensor]) -> torch.Tensor:
        """Defines the computation performed at every call.

        Args:
            inputs (List[torch.Tensor]): Inputs from the previous layers.

        Returns:
            (torch.Tensor): The output tensor of the block.
        """
        out = self.conv(torch.cat(inputs, 1))
        out = self.bn(out)
        out = self.relu(out)

        return out


class Tree(nn.Module):  # pylint: disable=too-many-instance-attributes
    """Tree node used to construct the structure of Hierarchical Deep
    Aggregation (HDA).

    Args;
        level (int): The stage number of this node in HDA.
        block (BasicBlock): The type of residual block used.
        in_channels (int): Number of channels in the input image.
        out_channels (int): Number of channels produced by the block.
        stride (int): Stride of the convolution.
        level_root (bool): Flag to indicate is this is the root node for the
            stage.
        dilation (int): Spacing between kernel elements. Default is 1.
        root_dim (int): Number of channels in the input image to the Root node.
        root_kernel_size (int): Size of the convolving kernel in the Root node.
        residual_root (bool): Flag to indicate if a residual layer should be
            used in the Root node. Default is False.
    """

    # pylint: disable=redefined-builtin

    def __init__(  # pylint: disable=too-many-arguments
        self,
        level: int,
        block: Callable[..., BasicBlock],
        in_channels: int,
        out_channels: int,
        stride: int = 1,
        level_root: bool = False,
        dilation: int = 1,
        root_dim: int = 0,
        root_kernel_size: int = 1,
    ) -> None:
        super().__init__()
        self.tree1: Union[BasicBlock, Tree]
        self.tree2: Union[BasicBlock, Tree]
        if root_dim == 0:
            root_dim = 2 * out_channels
        if level_root:
            root_dim += in_channels
        if level == 1:
            self.tree1 = block(in_channels, out_channels, stride, dilation=dilation)
            self.tree2 = block(out_channels, out_channels, 1, dilation=dilation)
        else:
            self.tree1 = Tree(
                level - 1,
                block,
                in_channels,
                out_channels,
                stride,
                dilation=dilation,
                root_dim=0,
                root_kernel_size=root_kernel_size,
            )
            self.tree2 = Tree(
                level - 1,
                block,
                out_channels,
                out_channels,
                dilation=dilation,
                root_dim=root_dim + out_channels,
                root_kernel_size=root_kernel_size,
            )
        if level == 1:
            self.root = Root(root_dim, out_channels, root_kernel_size)
        self.level_root = level_root
        self.root_dim = root_dim
        self.downsample = None
        self.project = None
        self.levels = level
        if stride > 1:
            self.downsample = nn.MaxPool2d(stride, stride=stride)
        if in_channels != out_channels:
            self.project = nn.Sequential(
                nn.Conv2d(
                    in_channels, out_channels, kernel_size=1, stride=1, bias=False
                ),
                nn.BatchNorm2d(out_channels, momentum=BN_MOMENTUM),
            )

    def forward(
        self,
        input: torch.Tensor,
        residual: torch.Tensor = None,
        children: List[torch.Tensor] = None,
    ) -> torch.Tensor:
        """Defines the computation performed at every call.

        Args:
            input (torch.Tensor): Input from the previous layer.
            residual (torch.Tensor): Residual from the connected node.
            children (List[torch.Tensor]): Inputs from child nodes of the
                current node.

        Returns:
            (torch.Tensor): The output tensor of the block.
        """
        children = [] if children is None else children
        bottom = self.downsample(input) if self.downsample else input
        residual = self.project(bottom) if self.project else bottom
        if self.level_root:
            children.append(bottom)
        out_1 = self.tree1(input, residual)
        if self.levels == 1:
            out_2 = self.tree2(out_1)
            out = self.root([out_2, out_1] + children)
        else:
            children.append(out_1)
            out = self.tree2(out_1, children=children)

        return out



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Counts the number of detected objects within a boundary."""

from typing import Any, Dict, List

from peekingduck.pipeline.nodes.dabble.zoningv1.zone import Zone
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Uses the bottom midpoints of all detected bounding boxes and outputs the
    number of object counts in each specified zone.

    Given the bottom mid-points of all detected objects, this node checks if
    the points fall within the area of the specified zones. The zone counting
    detections depend on the configuration set in the object detection models,
    such as the type of object to detect.

    Inputs:
        |btm_midpoint_data|

    Outputs:
        |zones_data|

        |zone_count_data|

    Configs:
        resolution (:obj:`List[int]`): **default = [1280, 720]**. |br|
            Resolution of input array to calculate pixel coordinates of zone
            points.
        zones (:obj:`List[List[List[Union[int, float]]]]`): |br|
            **default = [**                                     |br|
            |tab| **[[0, 0], [640, 0], [640, 720], [0, 720]],** |br|
            |tab| **[[0.5, 0], [1, 0], [1, 1], [0.5, 1]]**      |br|
            **]**                                               |br|
            Used for creation of specific zones with either the absolute pixel
            values or % of resolution as a fraction between [0, 1].
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        try:
            self.zones = [
                self._create_zone(zone, config["resolution"])  # type: ignore
                for zone in self.zones  # type: ignore
            ]
        except TypeError as error:
            self.logger.warning(error)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Counts all detected objects that falls within any specified zone,
        and return the total object count in each zone.
        """
        num_of_zones = len(self.zones)
        zone_counts = [0] * num_of_zones

        # for each x, y point, check if it is in any zone and add count
        for point in inputs["btm_midpoint"]:
            for i, zone in enumerate(self.zones):
                if zone.point_within_zone(*point):
                    zone_counts[i] += 1

        return {
            "zones": [zone.get_all_points_of_area() for zone in self.zones],
            "zone_count": zone_counts,
        }

    def _create_zone(self, zone: List[Any], resolution: List[int]) -> Any:
        """Creates the appropriate Zone given either the absolute pixel values
        or % of resolution as a fraction between [0, 1].
        """
        created_zone = None

        if all(all(0 <= i <= 1 for i in coords) for coords in zone):
            # coordinates are in fraction. Use resolution to get correct coords
            pixel_coords = [
                self._get_pixel_coords(coords, resolution) for coords in zone
            ]
            created_zone = Zone(pixel_coords)
        if all(all((isinstance(i, int) and i >= 0) for i in coords) for coords in zone):
            # when 1st-if fails and this statement passes, list is in pixel
            # value.
            created_zone = Zone(zone)

        # if neither, something is wrong
        if not created_zone:
            assert False, (
                "Zone %s needs to be all pixel-wise points or "
                "all fractions of the frame between 0 and 1. "
                "please check zone_count configs." % zone
            )

        return created_zone

    @staticmethod
    def _get_pixel_coords(coords: List[float], resolution: List[int]) -> List[float]:
        """Returns the pixel position of the zone points."""
        return [int(coords[0] * resolution[0]), int(coords[1] * resolution[1])]



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Blurs area bounded by bounding boxes over detected object.
"""

from typing import Any, Dict, List

import cv2
import numpy as np

from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):  # pylint: disable=too-few-public-methods
    """Blurs area bounded by bounding boxes on image.

    The :mod:`draw.blur_bbox` node blurs the areas of the image bounded by the
    bounding boxes output from an object detection model.

    Inputs:
        |img_data|

        |bboxes_data|

    Outputs:
        |img_data|

    Configs:
        blur_kernel_size (:obj:`int`): **default = 50**. |br|
            This defines the kernel size used in the blur filter. Larger values
            of ``blur_kernel_size`` gives more intense blurring.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

        self.blur_kernel_size = self.config["blur_kernel_size"]

    @staticmethod
    def _blur(
        bboxes: List[np.ndarray], image: np.ndarray, blur_kernel_size: int
    ) -> np.ndarray:
        """Blurs the area bounded by bbox in an image."""
        height = image.shape[0]
        width = image.shape[1]

        for bbox in bboxes:
            x_1, y_1, x_2, y_2 = bbox
            y_1, y_2 = int(y_1 * height), int(y_2 * height)
            x_1, x_2 = int(x_1 * width), int(x_2 * width)

            # to get the area bounded by bbox
            bbox_image = image[y_1:y_2, x_1:x_2, :]

            # apply the blur using blur filter from opencv
            blur_bbox_image = cv2.blur(bbox_image, (blur_kernel_size, blur_kernel_size))
            image[y_1:y_2, x_1:x_2, :] = blur_bbox_image

        return image

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Reads the image input and returns the image, with the areas bounded
        by the bboxes blurred.

        Args:
            inputs (dict): Dictionary of inputs with keys "img", "bboxes"

        Returns:
            outputs (dict): Output in dictionary format with key "img"
        """
        blurred_img = self._blur(inputs["bboxes"], inputs["img"], self.blur_kernel_size)
        outputs = {"img": blurred_img}
        return outputs



# Modifications copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Original copyright (c) 2019 ZhongdaoWang
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""JDE Multi-object Tracker.

Modifications include:
- Rearranged comments to they appear before the relevant lines of code
- Refactor variable names in update() for clarity
- Refactor subtract_stracks() to use list comprehension
- Refactor combine_stracks() to use bool for dictionary values instead
"""

import logging
from pathlib import Path
from typing import Any, Dict, List, Tuple

import numpy as np
import torch

from peekingduck.pipeline.nodes.model.jdev1.jde_files import matching
from peekingduck.pipeline.nodes.model.jdev1.jde_files.darknet import Darknet
from peekingduck.pipeline.nodes.model.jdev1.jde_files.kalman_filter import KalmanFilter
from peekingduck.pipeline.nodes.model.jdev1.jde_files.track import STrack, TrackState
from peekingduck.pipeline.nodes.model.jdev1.jde_files.utils import (
    letterbox,
    non_max_suppression,
    scale_coords,
)
from peekingduck.pipeline.utils.bbox.transforms import tlwh2xyxyn, xyxy2tlwh


class Tracker:  # pylint: disable=too-many-instance-attributes
    """JDE Multi-object Tracker.

    Args:
        config (Dict[str, Any]): Model configuration options.
        model_dir (Path): Directory to model weights files.
        frame_rate (float): Frame rate of the current video sequence, used
            for computing size of track buffer.
    """

    def __init__(
        self, config: Dict[str, Any], model_dir: Path, frame_rate: float
    ) -> None:
        self.logger = logging.getLogger(__name__)

        self.config = config
        self.device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        self.model = self._create_darknet_model(model_dir)

        self.tracked_stracks: List[STrack] = []
        self.lost_stracks: List[STrack] = []
        self.removed_stracks: List[STrack] = []

        self.frame_id = 0
        self.max_time_lost = int(frame_rate / 30.0 * config["track_buffer"])

        self.kalman_filter = KalmanFilter()

    def track_objects_from_image(
        self, image: np.ndarray
    ) -> Tuple[List[np.ndarray], List[int], List[float]]:
        """Tracks detections from the current video frame.

        Args:
            image (np.ndarray): The current video frame.

        Returns:
            (Tuple[List[np.ndarray], List[int], List[float]]): A tuple of
            - Numpy array of detected bounding boxes.
            - List of track IDs.
            - List of detection confidence scores.
        """
        image_size = image.shape[:2]
        padded_image = self._preprocess(image)
        padded_image = torch.from_numpy(padded_image).to(self.device).unsqueeze(0)

        online_targets = self.update(padded_image, image)
        online_tlwhs = []
        online_ids = []
        scores = []
        for target in online_targets:
            tlwh = target.tlwh
            vertical = tlwh[2] / tlwh[3] > 1.6
            if not vertical and tlwh[2] * tlwh[3] > self.config["min_box_area"]:
                online_tlwhs.append(tlwh)
                online_ids.append(target.track_id)
                scores.append(target.score.item())
        if not online_tlwhs:
            return online_tlwhs, online_ids, scores

        bboxes = self._postprocess(np.asarray(online_tlwhs), image_size)

        return bboxes, online_ids, scores

    @torch.no_grad()
    def update(  # pylint: disable=too-many-branches,too-many-locals,too-many-statements
        self, padded_image: torch.Tensor, image: np.ndarray
    ) -> List[STrack]:
        """Processes the image frame and finds bounding box (detections).

        Associates the detection with corresponding tracklets and also handles
        lost, removed, re-found and active tracklets

        Args:
            padded_image (torch.Tensor): Preprocessed image with letterbox
                resizing and color normalization.
            image (np.ndarray): The original video frame.

        Returns:
            (List[STrack]): The list contains information regarding the
            online tracklets for the received image tensor.
        """
        self.frame_id += 1
        # For storing active tracks, for the current frame
        activated_stracks = []
        # Lost Tracks whose detections are obtained in the current frame
        refind_stracks = []
        # The tracks which are not obtained in the current frame but are not
        # removed. (Lost for some time less than the threshold for removing)
        lost_stracks = []
        removed_stracks = []

        # Step 1: Network forward, get detections & embeddings
        # pred is tensor of all the proposals (default number of proposals:
        # 54264). Proposals have information associated with the bounding box
        # and embeddings
        pred = self.model(padded_image)
        # Proposals rejected on basis of object confidence score
        pred = pred[pred[..., 4] > self.config["score_threshold"]]
        if len(pred) > 0:
            # Final proposals are obtained in dets. Information of bounding box
            # and embeddings also included
            dets = non_max_suppression(
                pred.unsqueeze(0),
                self.config["nms_threshold"],
            )[0].cpu()
            # Next step changes the detection scales
            scale_coords(self.input_size, dets[:, :4], image.shape[:2]).round()

            # Detections is list of (x1, y1, x2, y2, object_conf, class_score,
            # class_pred) class_pred is the embeddings.
            detections = [
                STrack(xyxy2tlwh(xyxys[:4]), xyxys[4], f.numpy())
                for (xyxys, f) in zip(dets[:, :5], dets[:, 6:])
            ]
        else:
            detections = []

        # Add newly detected tracklets to tracked_stracks
        unconfirmed = []
        tracked_stracks: List[STrack] = []
        for track in self.tracked_stracks:
            if track.is_activated:
                # Active tracks are added to the local list 'tracked_stracks'
                tracked_stracks.append(track)
            else:
                # previous tracks which are not active in the current frame
                # are added in unconfirmed list
                unconfirmed.append(track)

        # Step 2: First association, with embedding
        # Combining currently tracked_stracks and lost_stracks
        strack_pool = _combine_stracks(tracked_stracks, self.lost_stracks)
        # Predict the current location with KF
        STrack.multi_predict(strack_pool, self.kalman_filter)

        # The dists is a matrix of distances of the detection with the tracks
        # in strack_pool
        dists = matching.embedding_distance(strack_pool, detections)
        dists = matching.fuse_motion(self.kalman_filter, dists, strack_pool, detections)
        # matches is the array for corresponding matches of the detection
        # with the corresponding strack_pool
        (
            matches,
            unmatched_track_indices,
            unmatched_det_indices,
        ) = matching.linear_assignment(dists, threshold=0.7)

        for tracked_idx, det_idx in matches:
            # tracked_idx is the id of the track and det_idx is the detection
            track = strack_pool[tracked_idx]
            det = detections[det_idx]
            if track.state == TrackState.TRACKED:
                # If the track is active, add the detection to the track
                track.update(detections[det_idx], self.frame_id)
                activated_stracks.append(track)
            else:
                # We have obtained a detection from a track which is not
                # active, hence put the track in refind_stracks list
                track.re_activate(det, self.frame_id)
                refind_stracks.append(track)

        # None of the steps below happen if there are no undetected tracks.
        # Step 3: Second association, with IOU
        # detections is now a list of the unmatched detections
        detections = [detections[i] for i in unmatched_det_indices]
        # This is container for stracks which were tracked till the
        # previous frame but no detection was found for it in the current frame
        r_tracked_stracks = []
        for i in unmatched_track_indices:
            if strack_pool[i].state == TrackState.TRACKED:
                r_tracked_stracks.append(strack_pool[i])
        dists = matching.iou_distance(r_tracked_stracks, detections)
        # matches is the list of detections which matched with corresponding
        # tracks by IOU distance method
        (
            matches,
            unmatched_track_indices,
            unmatched_det_indices,
        ) = matching.linear_assignment(dists, threshold=0.5)
        # Same process done for some unmatched detections, but now considering
        # IOU_distance as measure
        for tracked_idx, det_idx in matches:
            track = r_tracked_stracks[tracked_idx]
            det = detections[det_idx]
            if track.state == TrackState.TRACKED:
                track.update(det, self.frame_id)
                activated_stracks.append(track)
            else:  # pragma: no cover
                # This shouldn't be reached, r_tracked_stracks only takes in
                # tracks with TrackState.TRACKED from above
                track.re_activate(det, self.frame_id)
                refind_stracks.append(track)
        # If no detections are obtained for tracks (unmatched_track_indices),
        # the tracks are added to lost_tracks and are marked lost
        for i in unmatched_track_indices:
            track = r_tracked_stracks[i]
            if track.state != TrackState.LOST:
                track.mark_lost()
                lost_stracks.append(track)

        # Deal with unconfirmed tracks, usually tracks with only one beginning
        # frame
        detections = [detections[i] for i in unmatched_det_indices]
        dists = matching.iou_distance(unconfirmed, detections)
        (
            matches,
            unconfirmed_track_indices,
            unmatched_det_indices,
        ) = matching.linear_assignment(dists, threshold=0.7)
        for tracked_idx, det_idx in matches:
            unconfirmed[tracked_idx].update(detections[det_idx], self.frame_id)
            activated_stracks.append(unconfirmed[tracked_idx])
        # The tracks which are yet not matched
        for i in unconfirmed_track_indices:
            track = unconfirmed[i]
            track.mark_removed()
            removed_stracks.append(track)
        # after all these confirmation steps, if a new detection is found, it
        # is initialized for a new track
        # Step 4: Init new stracks
        for i in unmatched_det_indices:
            track = detections[i]
            if track.score < self.config["score_threshold"]:  # pragma: no cover
                # This shouldn't be reached since we already rejected proposals
                # on basis of object confidence score earlier
                continue
            track.activate(self.kalman_filter, self.frame_id)
            activated_stracks.append(track)

        # Step 5: Update state
        # If the tracks are lost for more frames than the threshold number, the
        # tracks are removed.
        for track in self.lost_stracks:
            if self.frame_id - track.end_frame > self.max_time_lost:
                track.mark_removed()
                removed_stracks.append(track)

        # Update the self.tracked_stracks and self.lost_stracks using the
        # updates in this step.
        self.tracked_stracks = [
            t for t in self.tracked_stracks if t.state == TrackState.TRACKED
        ]
        self.tracked_stracks = _combine_stracks(self.tracked_stracks, activated_stracks)
        self.tracked_stracks = _combine_stracks(self.tracked_stracks, refind_stracks)
        self.lost_stracks = _subtract_stracks(self.lost_stracks, self.tracked_stracks)
        self.lost_stracks.extend(lost_stracks)
        self.lost_stracks = _subtract_stracks(self.lost_stracks, self.removed_stracks)
        self.removed_stracks.extend(removed_stracks)
        self.tracked_stracks, self.lost_stracks = _remove_duplicate_stracks(
            self.tracked_stracks, self.lost_stracks
        )

        # get scores of lost tracks
        output_stracks = [track for track in self.tracked_stracks if track.is_activated]

        return output_stracks

    def _create_darknet_model(self, model_dir: Path) -> Darknet:
        """Creates a Darknet-53 model corresponding specified `model_type`.

        Args:
            model_dir (Path): Directory containing model weights files and
                backbone configuration files.

        Returns:
            (Darknet): Darknet backbone of the specified architecture and
                weights.
        """
        model_type = self.config["model_type"]
        model_path = model_dir / self.config["weights"]["model_file"][model_type]
        model_settings = self._parse_model_config(
            model_dir / self.config["weights"]["config_file"][model_type]
        )
        self.input_size = [
            int(model_settings[0]["width"]),
            int(model_settings[0]["height"]),
        ]
        self.logger.info(
            "JDE model loaded with the following configs:\n\t"
            f"Model type: {self.config['model_type']}\n\t"
            f"Input resolution: {self.input_size}\n\t"
            f"IOU threshold: {self.config['iou_threshold']}\n\t"
            f"NMS threshold: {self.config['nms_threshold']}\n\t"
            f"Score threshold: {self.config['score_threshold']}\n\t"
            f"Min bounding box area: {self.config['min_box_area']}\n\t"
            f"Track buffer: {self.config['track_buffer']}"
        )
        return self._load_darknet_weights(model_path, model_settings)

    def _load_darknet_weights(
        self, model_path: Path, model_settings: List[Dict[str, Any]]
    ) -> Darknet:
        """Loads pretrained Darknet-53 weights.

        Args:
            model_path (Path): Path to weights file.
            model_settings (List[Dict[str, Any]]): Model architecture
                configurations.

        Returns:
            (Darknet): Darknet backbone of the specified architecture and
                weights.
        """
        if not model_path.is_file():
            raise ValueError(
                f"Model file does not exist. Please check that {model_path} exists."
            )
        ckpt = torch.load(str(model_path), map_location="cpu")
        model = Darknet(model_settings, self.device, num_identities=14455)
        model.load_state_dict(ckpt["model"], strict=False)
        model.to(self.device).eval()
        return model

    def _preprocess(self, image: np.ndarray) -> np.ndarray:
        """Preprocesses the input image by padded resizing with letterbox and
        normalising RGB values.

        Args:
            image (np.ndarray): Input video frame.

        Returns:
            (np.ndarray): Preprocessed image.
        """
        # Padded resize
        padded_image = letterbox(
            image, height=self.input_size[1], width=self.input_size[0]
        )
        # Normalize RGB
        padded_image = padded_image[..., ::-1].transpose(2, 0, 1)
        padded_image = np.ascontiguousarray(padded_image, dtype=np.float32)
        padded_image /= 255.0

        return padded_image

    @staticmethod
    def _parse_model_config(config_path: Path) -> List[Dict[str, Any]]:
        """Parse model configuration. Currently parses all values to string.

        Args:
            config_path (Path): Path to model configuration file.

        Returns:
            (List[Dict[str, Any]]): A list of dictionaries each containing
                the configuration of a layer/module.
        """
        with open(config_path) as infile:
            lines = [
                line
                for line in map(str.strip, infile.readlines())
                if line and not line.startswith("#")
            ]
        module_defs: List[Dict[str, Any]] = []
        for line in lines:
            if line.startswith("["):
                module_defs.append({})
                module_defs[-1]["type"] = line[1:-1].rstrip()
                if module_defs[-1]["type"] == "convolutional":
                    module_defs[-1]["batch_normalize"] = 0
            else:
                key, value = tuple(map(str.strip, line.split("=")))
                if value.startswith("$"):
                    value = module_defs[0].get(value.strip("$"), None)
                module_defs[-1][key] = value
        return module_defs

    @staticmethod
    def _postprocess(tlwhs: np.ndarray, image_shape: Tuple[int, ...]) -> np.ndarray:
        """Post-processes detection bounding boxes by converting them from
        [t, l, w, h] to normalized [x1, y1, x2, y2] format which is required by
        other PeekingDuck draw nodes. (t, l) is the top-left corner, w is
        width, and h is height. (x1, y1) is the top-left corner and (x2, y2) is
        the bottom-right corner.

        Args:
            tlwhs (np.ndarray): Bounding boxes in [t, l, w, h] format.
            image_shape (Tuple[int, ...]): Dimensions of the original video
                frame.

        Returns:
            (np.ndarray): Bounding boxes in normalized [x1, y1, x2, y2] format.
        """
        return tlwh2xyxyn(tlwhs, *image_shape)


def _combine_stracks(stracks_1: List[STrack], stracks_2: List[STrack]) -> List[STrack]:
    """Combines two list of STrack together.

    Args:
        stracks_1 (List[STrack]): List of STrack.
        stracks_2 (List[STrack]): List of STrack.

    Returns:
        (List[STrack]): Combined list of STrack.
    """
    stracks = {track.track_id: track for track in stracks_1}
    for track in stracks_2:
        tid = track.track_id
        if tid not in stracks:
            stracks[tid] = track
    return list(stracks.values())


def _remove_duplicate_stracks(
    stracks_1: List[STrack], stracks_2: List[STrack]
) -> Tuple[List[STrack], List[STrack]]:
    """Remove duplicate STrack based on costs computed using
    Intersection-over-Union (IoU) values. Duplicates are identified by
    cost<0.15, the STrack that is more recently created is marked as
    the duplicate.

    Args:
        stracks_1 (List[STrack]): List of STrack.
        stracks_2 (List[STrack]): List of STrack.

    Returns:
        (Tuple[List[STrack], List[STrack]]): Lists of STrack with duplicates
            removed.
    """
    distances = matching.iou_distance(stracks_1, stracks_2)
    pairs = np.where(distances < 0.15)
    duplicates_1 = []
    duplicates_2 = []
    for idx_1, idx_2 in zip(*pairs):
        age_1 = stracks_1[idx_1].frame_id - stracks_1[idx_1].start_frame
        age_2 = stracks_2[idx_2].frame_id - stracks_2[idx_2].start_frame
        if age_1 > age_2:
            duplicates_2.append(idx_2)
        else:
            duplicates_1.append(idx_1)
    return (
        [t for i, t in enumerate(stracks_1) if i not in duplicates_1],
        [t for i, t in enumerate(stracks_2) if i not in duplicates_2],
    )


def _subtract_stracks(stracks_1: List[STrack], stracks_2: List[STrack]) -> List[STrack]:
    """Removes stracks_2 from stracks_1.

    Args:
        stracks_1 (List[STrack]): List of STrack.
        stracks_2 (List[STrack]): List of STrack.

    Returns:
        (List[STrack]): List of STrack.
    """
    stracks = {track.track_id: track for track in stracks_1}
    for track in stracks_2:
        tid = track.track_id
        if tid in stracks:
            del stracks[tid]
    return list(stracks.values())



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility functions for statistics
"""
import re
from typing import Any, Callable, Dict, List, Tuple, Union


class Stats:
    """
    Extract the chosen <unary_expr> or <cond_expr> config and apply to incoming data.

    Args:
        ops (Dict[str, Callable]): A dictionary of operator symbols (keys) pointing to
        callable operator modules (values).
    """

    def __init__(self, ops: Dict[str, Callable]) -> None:
        self.ops = ops
        self.func: str
        self.expr: str
        self.condition = {"op_func": Callable, "operand": str}
        self.data_type: str
        self.keys: List[str] = []
        self.func_map: Dict[str, Dict[str, Any]] = {
            "identity": {"types": (int, float), "func": _func_identity},
            "length": {"types": (dict, list), "func": _func_length},
            "minimum": {"types": (dict, list), "func": _func_minimum},
            "maximum": {"types": (dict, list), "func": _func_maximum},
            "cond_count": {"types": (list), "func": _func_cond_count},
        }

    def prepare_data(
        self, all_funcs: Dict[str, Union[str, None]]
    ) -> Tuple[str, List[str]]:
        """
        Does the necessary data preparation during node initialisation.

        Args:
            - all_funcs (Dict[str, Union[str, None]]): Dictionary of <func>: <user input>.

        Returns:
            - self.data_type (str): PeekingDuck in-built data type or custom data type from custom
            nodes.
            - self.keys (List[str]): List of selected keys if self.data_type is a dictionary.
            If not, returns an empty list.
        """
        self.func, self.expr = _get_func_expr(all_funcs)
        self._update_data_type_keys_condition()

        return self.data_type, self.keys

    def get_curr_result(self, inputs: Any, keys: List[str]) -> Union[float, int, None]:
        """
        Extracts the current resulting value from the incoming data.

        Args:
            - inputs (Any): Current value for the self.data_type key.
            - keys (List[str]): List of selected keys if self.data_type is a dictionary.

        Returns:
            - curr (Union[float, int, None]): Current resulting value.

        """
        target_attr = _deep_get_value(inputs, keys)
        curr = self._apply_func(target_attr)

        return curr

    def _update_data_type_keys_condition(self) -> None:
        """Uses self.func, self.expr, and self.ops to update self.data_type, self.keys, and
        self.condition."""
        ops_expr = "(" + ")|(".join(self.ops.keys()) + ")"
        match = re.search(ops_expr, self.expr)

        if not match and self.func == "cond_count":
            raise ValueError(
                f"The chosen function: {self.func} should have an operator for comparison."
            )
        if match and self.func != "cond_count":
            raise ValueError(
                f"The chosen function: {self.func} should not have the {match.group()} "
                f"operator."
            )

        if match:
            operator = match.group()
            self.condition["op_func"] = self.ops[operator]
            op_idx_start, op_idx_end = match.span()
            target_attr = self.expr[:op_idx_start]
            operand_raw = self.expr[op_idx_end:]
            self.condition["operand"] = _get_operand(operand_raw, operator)  # type: ignore
        else:
            target_attr = self.expr

        self.data_type, self.keys = _get_data_type_and_keys(target_attr)

    def _apply_func(self, target_attr: Any) -> Union[int, float, None]:
        """Applies a function and optional condition to a target attribute."""
        if not target_attr:
            return None

        _check_type(target_attr, self.func, self.func_map[self.func]["types"])
        args = (target_attr, self.condition)
        return self.func_map[self.func]["func"](*args)


def _get_func_expr(all_funcs: Dict[str, Union[str, None]]) -> Tuple[str, str]:
    """Gets the only non-null key value pair from a dictionary."""
    num_funcs = 0
    for func, expr in all_funcs.items():
        if expr:
            num_funcs += 1
            selected_func, selected_expr = func, expr

    if num_funcs < 1:
        raise ValueError(
            "No functions selected in config, but one function needs to be selected to proceed."
        )
    if num_funcs > 1:
        raise ValueError(
            f"{num_funcs} functions selected in config, but only one function should be selected "
            f"to proceed."
        )
    return selected_func, selected_expr


def _get_data_type_and_keys(target_attr: str) -> Tuple[str, List[str]]:
    """Extract the name of data_type and its associated keys in a list, if any."""
    target_attr = re.sub(r"'|\"| ", "", target_attr)
    keys = re.findall(r"\[(.*?)\]", target_attr)
    if not keys:
        data_type = target_attr
        keys = []
    else:
        data_type = target_attr.split("[")[0]

    return data_type, keys


def _get_operand(operand_raw: str, operator: str) -> Union[str, float]:
    """Removes leading and trailing spaces, and return as a string or float."""
    operand_raw = operand_raw.strip()
    is_string = re.findall("'|\"", operand_raw)
    if is_string:
        if operator != "==":
            raise ValueError(
                f"The detected operand is: {operand_raw} and detected operator is: {operator}. "
                f"The operand is assumed to be a string as it is enclosed by single or double "
                f"quotes, and for string operand, only the '==' operator should be used for "
                f"comparison."
            )
        operand = re.sub("'|\"", "", operand_raw)
        return operand
    try:
        return float(operand_raw)
    except ValueError as error:
        raise ValueError(
            f"The detected operand here is: {operand_raw}."
            f"If the operand is intended to be a string, ensure that it is enclosed by single "
            f"or double quotes. Else, it is assumed to be of integer or float type, and will be "
            f"subsequently converted into a float."
        ) from error


def _deep_get_value(
    data: Dict[str, Any], keys: List[str]
) -> Union[int, float, list, dict]:
    """Recursively goes through the keys of a dictionary to obtain the final value."""
    if not keys:
        return data
    key = keys.pop(0)
    return _deep_get_value(data[key], keys)


def _func_identity(target_attr: Any, *_: Any) -> Union[float, int]:
    """Function for returning the identity of an integer or float."""
    return target_attr


def _func_length(target_attr: Union[Dict[str, Any], List[Any]], *_: Any) -> int:
    """Function for returning the length of a dictionary or list."""
    return len(target_attr)


def _func_minimum(
    target_attr: Union[
        Dict[Union[float, int], Union[float, int]], List[Union[float, int]]
    ],
    *_: Any,
) -> Union[float, int]:
    """Function for returning the minimum element of a list or dictionary."""
    try:
        return min(target_attr)
    except ValueError as error:
        raise ValueError(
            "To use the 'minimum' function, all elements within target_attr has to be"
            " of type 'int' or 'float'."
        ) from error


def _func_maximum(
    target_attr: Union[
        Dict[Union[float, int], Union[float, int]], List[Union[float, int]]
    ],
    *_: Any,
) -> Union[float, int]:
    """Function for returning the maximum element of a list or dictionary."""
    try:
        return max(target_attr)
    except ValueError as error:
        raise ValueError(
            "To use the 'maximum' function, all elements within the target_attr has to be"
            " of type 'int' or 'float'."
        ) from error


def _func_cond_count(target_attr: Any, condition: Dict[str, Any]) -> int:
    """Function for counting the number of elements in a list, given an operator and operand
    for comparison."""
    count = 0
    for item in target_attr:
        if condition["op_func"](item, condition["operand"]):
            count += 1
    return count


def _check_type(
    target_attr: Union[int, float, list, dict], func: str, *types: type
) -> None:
    """Checks that a target attribute conforms to the defined type(s)."""
    correct = False
    for obj_type in types:
        if isinstance(target_attr, obj_type):
            correct = True
    if not correct:
        raise TypeError(
            f"For the chosen function: '{func}', valid target_attr types are: {types}. "
            f"However, this target_attr: {target_attr} is of type: {type(target_attr)}."
        )



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Calculates the cumulative average, minimum, and maximum of a single variable
of interest over time.
"""

import operator
from typing import Any, Dict, Union

from peekingduck.pipeline.nodes.dabble.statisticsv1 import utils
from peekingduck.pipeline.nodes.node import AbstractNode

# Order matters so that regex doesn't read ">=" as ">" or "<=" as "<"
# Dictionaries are insertion ordered from Python 3.6 onwards
OPS = {
    ">=": operator.ge,
    ">": operator.gt,
    "==": operator.eq,
    "<=": operator.le,
    "<": operator.lt,
}


class Node(AbstractNode):  # pylint: disable=too-many-instance-attributes
    """Calculates the cumulative average, minimum, and maximum of a single
    variable of interest (defined as ``current result`` here) over time. The
    configurations for this node offer several functions to reduce the incoming
    data type into a single ``current result`` of type :obj:`int` or
    :obj:`float`, which is valid for the current video frame.
    ``current result`` is then used to recalculate the values of the cumulative
    average, minimum, and maximum for PeekingDuck's running duration thus far.

    The configuration for this node is described below using a combination of
    the `Extended BNF
    <https://en.wikipedia.org/wiki/Extended_Backus%E2%80%93Naur_form>`_ and
    `Augmented BNF
    <https://en.wikipedia.org/wiki/Augmented_Backus%E2%80%93Naur_form>`_
    metasyntax. Concrete examples are provided later for illustration. ::

        pkd_data_type   = ? PeekingDuck built-in data types ?
                          e.g. count, large_groups, obj_attrs
        user_data_type  = ? user data types produced by custom nodes ?
                          e.g. my_var, my_attrs
        dict_key        = ? Python dictionary keys, with optional nesting ?
                          e.g. ["ids"], ["details"]["age"]
        data_type       = pkd_data_type | user_data_type
        target_attr     = data_type | data_type "[" dict_key "]"

        unary_function  = "identity" | "length" | "maximum" | "minimum"
        unary_expr      = unary_function ":" target_attr

        num_operator    = "==" | ">=" | "<=" | ">" | "<"
        num_operand     = ? Python integers or floats ?
        num_comparison  = num_operator num_operand

        str_operator    = "=="
        str_operand     = ? Python strings enclosed by single or double quotes ?
        str_comparison  = str_operator str_operand

        cond_function   = "cond_count"
        cond_expr       = cond_function ":" target_attr ( num_comparison | str_comparison )

        configuration   = unary_expr | cond_expr

    Points to note:

        * Square brackets (``[]``) are used to define ``<dict_key>``, and
          should not be used elsewhere in the configuration.
        * Operands are processed differently depending on whether they are
          enclosed by single/double quotes, or not. If enclosed, the operand is
          assumed to be of type :obj:`str` and classified as ``<str_operand>``.
          If not, the operand is classified as ``<num_operand>`` and converted
          into :obj:`float` for further processing.

    The table below illustrates how configuration choices reduce the incoming
    data type into the ``<current result>``.

    +---------------------------------------+-------------------+-------------------+-------------+
    | ``<pkd_data_type>``: value            | ``<target_attr>`` | ``<unary_expr>``  | ``<current  |
    |                                       |                   |                   | result>``   |
    | or                                    |                   | or                |             |
    |                                       |                   |                   |             |
    | ``<user_data_type>``: value           |                   | ``<cond_expr>``   |             |
    +---------------------------------------+-------------------+-------------------+-------------+
    | count: 8                              | count             | identity:         | 8           |
    |                                       |                   |                   |             |
    |                                       |                   | count             |             |
    +---------------------------------------+-------------------+-------------------+-------------+
    | obj_attrs: {                          | obj_attrs["ids"]  | length:           | 3           |
    |                                       |                   |                   |             |
    |                                       |                   | obj_attrs["ids"]  |             |
    |   ids: [1,2,4],                       +-------------------+-------------------+-------------+
    |                                       | obj_attrs         | maximum:          | 52          |
    |   details: {                          | ["details"]       |                   |             |
    |                                       | ["age"]           | obj_attrs         |             |
    |     gender: ["male","male","female"], |                   | ["details"]       |             |
    |                                       |                   | ["age"]           |             |
    |     age: [52,17,48] }}                +-------------------+-------------------+-------------+
    |                                       | obj_attrs         | cond_count:       | 2           |
    |                                       | ["details"]       |                   |             |
    |                                       | ["gender"]        | obj_attrs         |             |
    |                                       |                   | ["details"]       |             |
    |                                       |                   | ["gender"]        |             |
    |                                       |                   |                   |             |
    |                                       |                   | == "male"         |             |
    |                                       +-------------------+-------------------+-------------+
    |                                       | obj_attrs         | cond_count:       | 3           |
    |                                       | ["details"]       |                   |             |
    |                                       | ["age"]           | obj_attrs         |             |
    |                                       |                   | ["details"]       |             |
    |                                       |                   | ["age"]           |             |
    |                                       |                   |                   |             |
    |                                       |                   | < 60              |             |
    +---------------------------------------+-------------------+-------------------+-------------+

    Inputs:
        |all_input_data|

    Outputs:
        |cum_avg_data|

        Note that :term:`cum_avg` will not be updated if there are no
        detections. For example, if :term:`cum_avg` = 10 for video frame 1, and
        there are no detections in the following 500 frames, :term:`cum_avg` is
        still 10 for video frame 501.

        |cum_max_data|

        |cum_min_data|

    Configs:
        identity (:obj:`str`): **default=null** |br|
            Accepts ``<target_attr>`` of types :obj:`int` or :obj:`float`, and
            returns the same value.
        length (:obj:`str`): **default=null** |br|
            Accepts ``<target_attr>`` of types :obj:`List[Any]` or
            :obj:`Dict[str, Any]`, and returns its length.
        minimum (:obj:`str`): **default=null** |br|
            Accepts ``<target_attr>`` of types :obj:`List[float | int]` or
            :obj:`Dict[str, float | int]`, and returns the minimum element
            within for the current frame. Not to be confused with the
            :term:`cum_min` output data type, which represents the cumulative
            minimum over time.
        maximum (:obj:`str`): **default=null** |br|
            Accepts ``<target_attr>`` of types :obj:`List[float | int]` or
            :obj:`Dict[str, float | int]`, and returns the maximum element
            within for the current frame. Not to be confused with the
            :term:`cum_max` output data type, which represents the cumulative
            maximum over time.
        cond_count (:obj:`str`): **default=null** |br|
            Accepts ``<target_attr>`` of types :obj:`List[float | int | str]`,
            and checks if each element in the list fulfils the condition
            described by ``<num_comparison>`` or ``<str_comparison>``. The
            number of elements that fulfil the condition are counted towards
            ``<current result>``.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.cum_avg, self.cum_min, self.cum_max = 0.0, float("inf"), float("-inf")
        self.num_iter = 0
        all_funcs = {
            "cond_count": self.cond_count,
            "identity": self.identity,
            "length": self.length,
            "minimum": self.minimum,
            "maximum": self.maximum,
        }
        self.stats = utils.Stats(OPS)
        self.data_type, self.keys = self.stats.prepare_data(all_funcs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Calculates the average, minimum and maximum of a single variable of interest over time.

        Args:
            inputs (dict): Dictionary with all available keys.

        Returns:
            outputs (dict): Dictionary with keys "cum_avg", "cum_min" and "cum_max".
        """

        self.curr = self.stats.get_curr_result(inputs[self.data_type], self.keys.copy())

        # if no detections in this frame, do not update and return stats from previous detections
        if self.curr:
            self._update_stats(self.curr)

        return {
            "cum_avg": self.cum_avg,
            "cum_min": self.cum_min,
            "cum_max": self.cum_max,
        }

    def _update_stats(self, curr: Union[float, int]) -> None:
        """Updates the cum_avg, cum_min and cum_max values with the current value."""
        if not isinstance(curr, float) and not isinstance(curr, int):
            raise TypeError(
                f"The current result has to be of type 'int' or 'float' to calculate statistics."
                f"However, the current result here is: '{curr}' which is of type: {type(curr)}."
            )

        if curr < self.cum_min:
            self.cum_min = curr
        if curr > self.cum_max:
            self.cum_max = curr
        if self.num_iter == 0:
            self.cum_avg = curr
        else:
            self.cum_avg = (self.cum_avg * self.num_iter + curr) / (self.num_iter + 1)
        self.num_iter += 1



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Draws bounding boxes over detected objects.
"""

from typing import Any, Dict

from peekingduck.pipeline.nodes.draw.utils.bbox import draw_bboxes
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Draws bounding boxes on image.

    The :mod:`draw.bbox` node uses :term:`bboxes` and, optionally,
    :term:`bbox_labels` from the model predictions to draw the bbox predictions
    onto the image.

    Inputs:
        |img_data|

        |bboxes_data|

        |bbox_labels_data|

    Outputs:
        |none_output_data|

    Configs:
        show_labels (:obj:`bool`): **default = False**. |br|
            If ``True``, shows class label, e.g., "person", above the bounding
            box.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        draw_bboxes(
            inputs["img"], inputs["bboxes"], inputs["bbox_labels"], self.show_labels
        )
        return {}



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Draws a tag (from :term:`obj_attrs`) above each bounding box."""

import copy
from typing import Any, Dict, List

from peekingduck.pipeline.nodes.draw.utils.bbox import check_bgr_type, draw_tags
from peekingduck.pipeline.nodes.node import AbstractNode


class Node(AbstractNode):
    """Draws a tag above each bounding box in the image, using information from
    selected attributes in :term:`obj_attrs`. In the general example below,
    :term:`obj_attrs` has 2 attributes (`<attr a>` and `<attr b>`). There are
    `n` detected bounding boxes, and each attribute has `n` corresponding tags
    stored in a list. The ``show`` config described subsequently is used to
    choose the attribute or attributes to be drawn. ::

        {"obj_attrs": {<attr a>: [<tag 1>, ..., <tag n>], <attr b>: [<tag 1>, ..., <tag n>]}}

    The following type conventions need to be observed:

        * Each attribute must be of type :obj:`List`, e.g.,
          ``<attr a>: [<tag 1>, ..., <tag n>]``
        * Each tag must be of type :obj:`str`, :obj:`int`, :obj:`float`, or
          :obj:`bool` to be convertable into :obj:`str` type for drawing

    In the example below, :term:`obj_attrs` has 3 attributes (`"ids"`,
    `"gender"` and `"age"`), where the last 2 attributes are nested within
    `"details"`. There are 2 detected bounding boxes, and thus each attribute
    consists of a list with 2 tags. ::

        # Example
        {"obj_attrs": {"ids":[1,2], "details": {"gender": ["female","male"], "age": [52,17]}}

    The table below illustrates how ``show`` can be configured to achieve
    different outcomes for this example. Key takeaways are:

        * To draw nested attributes, include all the keys leading to them
          (within the :term:`obj_attrs` dictionary), separating each key with a
          ``->``.
        * To draw multiple comma-separated attributes above each bounding box,
          add them to the list of ``show`` config.

    +-----+-----------------------------------------+---------------+---------------+
    | No. | ``show`` config                         | Tag above 1st | Tag above 2nd |
    |     |                                         | bounding box  | bounding box  |
    +=====+=========================================+===============+===============+
    | 1.  | ["ids"]                                 | "1"           | "2"           |
    +-----+-----------------------------------------+---------------+---------------+
    | 2.  | ["details -> gender"]                   | "female"      | "male"        |
    +-----+-----------------------------------------+---------------+---------------+
    | 3.  | ["details -> age", "details -> gender"] | "52, female"  | "17, male"    |
    +-----+-----------------------------------------+---------------+---------------+

    Inputs:
        |img_data|

        |bboxes_data|

        |obj_attrs_data|

    Outputs:
        |none_output_data|

    Configs:
        show (:obj:`List[str]`): **default = []**. |br|
            List of desired attributes to be drawn. For more details on how to
            use this config, see the section above.
        tag_color (:obj:`List[int]`): **default = [77, 103, 255]**. |br|
            Define the color of the drawn tag, in BGR format. Defined values
            have to be integers, and :math:`0 \\leq value \\leq 255`.

    .. versionchanged:: 1.2.0
        :mod:`draw.tag` used to take in ``obj_tags`` (:obj:`List[str]`) as an
        input data type, which has been deprecated and now subsumed under
        :term:`obj_attrs`, giving this node more flexibility. Also, the
        ``tag_color`` config is added to provide the option of changing the
        tag's color.
    """

    def __init__(self, config: Dict[str, Any] = None, **kwargs: Any) -> None:
        super().__init__(config, node_path=__name__, **kwargs)
        self.attr_keys = []
        check_bgr_type(self.tag_color)
        if not self.show:
            raise KeyError(
                "The 'show' config is currently empty. Add the desired attributes to be drawn "
                "to the list, in order to proceed."
            )
        for attr in self.show:
            attr = attr.replace(" ", "")
            self.attr_keys.append(attr.split(r"->"))

    def run(self, inputs: Dict[str, Any]) -> Dict[str, Any]:
        """Draws a tag above each bounding box.

        Args:
            inputs (dict): Dictionary with keys "bboxes", "obj_attrs", "img".

        Returns:
            outputs (dict): Dictionary with keys "none".
        """
        tags = self._tags_from_obj_attrs(inputs["obj_attrs"])
        # if empty list, nothing to draw
        if not tags:
            return {}
        draw_tags(inputs["img"], inputs["bboxes"], tags, self.tag_color)

        return {}

    def _tags_from_obj_attrs(self, inputs: Dict[str, Any]) -> List[str]:
        """Process inputs from various attributes into tags for drawing."""
        all_attrs: List[List[Any]] = []
        attr_keys = copy.deepcopy(self.attr_keys)
        for attr_key in attr_keys:
            attr = _deep_get_value(inputs, attr_key)
            if not isinstance(attr, list):
                raise TypeError(
                    f"The attribute of interest has to be of type 'list', containing a list of "
                    f"tags. However, the attribute chosen here was: {attr} which is of type: "
                    f"{type(attr)}."
                )
            all_attrs.append(attr)

        tags = []
        # all_attrs: [["a","b"], [1,2]] -> list(zip): [("a",1), ("b",2)] -> tags: ["a, 1", "b, 2"]
        for idx, obj in enumerate(list(zip(*all_attrs))):
            # only check for the first object for efficiency
            if idx == 0:
                for tag in obj:
                    if not _check_valid_type(tag, str, int, float, bool):
                        raise TypeError(
                            f"A tag has to be of type 'str', 'int', 'float' or 'bool' to be "
                            f"convertable to a string. However, the tag: {tag} is of type: "
                            f"{type(tag)}"
                        )
            attr_str = map(str, obj)
            tags.append(", ".join(attr_str))

        return tags


def _deep_get_value(data: Dict[str, Any], keys: List[str]) -> Dict[str, Any]:
    """Recursively goes through the keys of a dictionary to obtain the final value."""
    if not keys:
        return data
    key = keys.pop(0)
    return _deep_get_value(data[key], keys)


def _check_valid_type(tag: Any, *types: type) -> bool:
    """Checks the type of tag against all given types."""
    for data_type in types:
        if isinstance(tag, data_type):
            return True
    return False



# Copyright 2022 AI Singapore
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Finds paths for where weights are stored.
"""

from pathlib import Path
from typing import Any, Dict, Tuple, Optional

PEEKINGDUCK_WEIGHTS_SUBDIR = "peekingduck_weights"


def find_paths(
    root: Path, weights: Dict[str, Any], weights_parent_dir: Optional[str] = None
) -> Tuple[Path, Path]:
    """Checks for model weight paths from weights folder.

    Args:
        root (:obj:`str`): Path of ``peekingduck`` root folder.
        weights (:obj:`Dict[str, Any]`): File/subdir names of weights.
        weights_parent_dir (:obj:`Optional[str]`): Parent dir of where weights will be stored.

    Returns:
        weights_dir (:obj:`pathlib.Path`): Path to where all weights are stored.
        model_dir (:obj:`pathlib.Path`): Path to where weights for a model are stored.
    """

    if weights_parent_dir is None:
        weights_dir = root.parent / PEEKINGDUCK_WEIGHTS_SUBDIR
    else:
        weights_parent_path = Path(weights_parent_dir)
        if not weights_parent_path.exists():
            raise FileNotFoundError(
                f"The specified weights_parent_dir: {weights_parent_dir} does not exist."
            )
        if not weights_parent_path.is_absolute():
            raise ValueError(
                f"The specified weights_parent_dir: {weights_parent_dir} "
                "has to be an absolute path."
            )
        weights_dir = weights_parent_path / PEEKINGDUCK_WEIGHTS_SUBDIR

    model_dir = weights_dir / weights["model_subdir"]

    return weights_dir, model_dir
